# -*- coding: utf-8 -*-

import abc

import numpy as np

from skyllh.core.config import (
    HasConfig,
)
from skyllh.core.datafields import (
    DataFields,
    DataFieldStages as DFS,
)
from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core.event_selection import (
    AllEventSelectionMethod,
    EventSelectionMethod,
)
from skyllh.core.py import (
    classname,
    float_cast,
    func_has_n_args,
    issequenceof,
)
from skyllh.core.random import (
    RandomChoice,
)
from skyllh.core.scrambling import (
    DataScrambler,
)
from skyllh.core.timing import (
    TaskTimer,
)


logger = get_logger(__name__)


class BackgroundGenerationMethod(
        HasConfig,
        metaclass=abc.ABCMeta,
):
    """This is the abstract base class for a detector specific background
    generation method.
    """

    def __init__(
            self,
            **kwargs,
    ):
        """Constructs a new background generation method instance.
        """
        super().__init__(**kwargs)

    def change_shg_mgr(self, shg_mgr):
        """Notifies the background generation method about an updated
        SourceHypoGroupManager instance.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The new instance of SourceHypoGroupManager.
        """
        pass

    @abc.abstractmethod
    def generate_events(
            self,
            rss,
            dataset,
            data,
            mean,
            tl=None,
            **kwargs,
    ):
        """This method is supposed to generate a `mean` number of background
        events for the given dataset and its data.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to generate
            random numbers from.
        dataset : instance of Dataset
            The Dataset instance describing the dataset for which background
            events should get generated.
        data : instance of DatasetData
            The DatasetData instance holding the data of the dataset for which
            background events should get generated.
        mean : float
            The mean number of background events to generate.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.
        **kwargs
            Additional keyword arguments, which might be required for a
            particular background generation method.

        Returns
        -------
        n_bkg : int
            The number of generated background events.
        bkg_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the generated
            background events. The number of events in this array might be less
            than ``n_bkg`` if an event selection method was used for
            optimization purposes. The difference ``n_bkg - len(bkg_events)`` is
            then the number of pure background events in the generated
            background event sample.
        """
        pass


class MCDataSamplingBkgGenMethod(
        BackgroundGenerationMethod,
):
    """This class implements the method to generate background events from
    monte-carlo (MC) data by sampling events from the MC data set according to a
    probability value given for each event. Functions can be provided to get the
    mean number of background events and the probability of each monte-carlo
    event.
    """
    def __init__(
            self,
            get_event_prob_func,
            get_mean_func=None,
            data_scrambler=None,
            keep_mc_data_fields=None,
            pre_event_selection_method=None,
            **kwargs,
    ):
        """Creates a new instance of the MCDataSamplingBkgGenMethod class.

        Parameters
        ----------
        get_event_prob_func : callable
            The function to get the background probability of each monte-carlo
            event. The call signature of this function must be

                __call__(dataset, data, events)

            where ``dataset`` is an instance of Dataset and ``data`` is an
            instance of DatasetData of the data set for which background events
            needs to get generated. The ``events`` argument holds the actual
            set of events, for which the background event probabilities need to
            get calculated.
        get_mean_func : callable | None
            The function to get the mean number of background events.
            The call signature of this function must be

                __call__(dataset, data, events)

            where ``dataset`` is an instance of Dataset and ``data`` is an
            instance of DatasetData of the data set for which background events
            needs to get generated. The `events` argument holds the actual set
            of events, for which the mean number of background events should get
            calculated. This argument can be `None`, which means that the mean
            number of background events to generate needs to get specified
            through the ``generate_events`` method. However, if a pre event
            selection method is provided, this argument cannot be ``None``!
        data_scrambler : instance of DataScrambler | None
            If set to an instance of DataScrambler, the drawn monte-carlo
            background events will get scrambled. This can ensure more
            independent data trials. It is especially important when monte-carlo
            statistics are low.
        keep_mc_data_fields : str | list of str | None
            The MC data field names that should be kept in order to be able to
            calculate the background events rates by the functions
            ``get_event_prob_func`` and ``get_mean_func``. All other MC fields
            will get dropped due to computational efficiency reasons.
        pre_event_selection_method : instance of EventSelectionMethod | None
            If set to an instance of EventSelectionMethod, this method will
            pre-select the MC events that will be used for later background
            event generation. Using this pre-selection a large portion of the
            MC data can be reduced prior to background event generation.
        """
        super().__init__(
            **kwargs)

        self.get_event_prob_func = get_event_prob_func
        self.get_mean_func = get_mean_func
        self.data_scrambler = data_scrambler
        self.keep_mc_data_field_names = keep_mc_data_fields
        self.pre_event_selection_method = pre_event_selection_method

        if (pre_event_selection_method is not None) and (get_mean_func is None):
            raise ValueError(
                'If an event pre-selection method is provided, a '
                'get_mean_func needs to be provided as well!')

        # Define cache members to cache the background probabilities for each
        # monte-carlo event. The probabilities change only if the data changes.
        self._cache_data_id = None
        self._cache_mc = None
        self._cache_mc_event_bkg_prob = None
        self._cache_mean = None
        self._cache_mean_pre_selected = None
        self._cache_random_choice = None

    @property
    def get_event_prob_func(self):
        """The function to obtain the background probability for each
        monte-carlo event of the data set.
        """
        return self._get_event_prob_func

    @get_event_prob_func.setter
    def get_event_prob_func(self, func):
        if not callable(func):
            raise TypeError(
                'The get_event_prob_func property must be a callable! '
                f'Its current type is {classname(func)}.')
        if not func_has_n_args(func, 3):
            raise TypeError(
                'The function provided for the get_event_prob_func property '
                'must have 3 arguments!')
        self._get_event_prob_func = func

    @property
    def get_mean_func(self):
        """The function to obtain the mean number of background events for the
        data set. This can be None, which means that the mean number of
        background events to generate needs to be specified through the
        `generate_events` method.
        """
        return self._get_mean_func

    @get_mean_func.setter
    def get_mean_func(self, func):
        if func is None:
            self._get_mean_func = None
            return

        if not callable(func):
            raise TypeError(
                'The get_mean_func property must be a callable! '
                f'Its current type is {classname(func)}.')
        if not func_has_n_args(func, 3):
            raise TypeError(
                'The function provided for the get_mean_func property must '
                'have 3 arguments!')
        self._get_mean_func = func

    @property
    def data_scrambler(self):
        """The DataScrambler instance that should be used to scramble the drawn
        monte-carlo background events to ensure more independent data trials.
        This is especially important when monte-carlo statistics are low. It is
        `None`, if no data scrambling should be used.
        """
        return self._data_scrambler

    @data_scrambler.setter
    def data_scrambler(self, scrambler):
        if scrambler is None:
            self._data_scrambler = None
            return

        if not isinstance(scrambler, DataScrambler):
            raise TypeError(
                'The data_scrambler property must be an instance of '
                'DataScrambler! '
                f'Its current type is {classname(scrambler)}.')
        self._data_scrambler = scrambler

    @property
    def keep_mc_data_field_names(self):
        """The MC data field names that should be kept in order to be able to
        calculate the background events rates by the functions
        ``get_event_prob_func`` and ``get_mean_func``. All other MC fields
        will get dropped due to computational efficiency reasons.
        """
        return self._keep_mc_data_field_names

    @keep_mc_data_field_names.setter
    def keep_mc_data_field_names(self, names):
        if names is None:
            names = []
        elif isinstance(names, str):
            names = [names]
        elif not issequenceof(names, str):
            raise TypeError(
                'The keep_mc_data_field_names must be None, an instance of '
                'type str, or a sequence of instances of type str!')
        self._keep_mc_data_field_names = names

    @property
    def pre_event_selection_method(self):
        """The instance of EventSelectionMethod that pre-selects the MC events,
        which can be considered for background event generation.
        """
        return self._pre_event_selection_method

    @pre_event_selection_method.setter
    def pre_event_selection_method(self, method):
        if method is None:
            self._pre_event_selection_method = None
            return

        if not isinstance(method, EventSelectionMethod):
            raise TypeError(
                'The pre_event_selection_method property must be None, or an '
                'instance of EventSelectionMethod!')

        # If the event selection method selects all events, it's equivalent
        # to have it set to None, because then no operation has to be
        # performed.
        if isinstance(method, AllEventSelectionMethod):
            method = None

        self._pre_event_selection_method = method

    def change_shg_mgr(self, shg_mgr):
        """Changes the instance of SourceHypoGroupManager of the
        pre-event-selection method. Also it invalidates the data cache of this
        background generation method.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The new instance of SourceHypoGroupManager.
        """
        if self._pre_event_selection_method is not None:
            self._pre_event_selection_method.change_shg_mgr(
                shg_mgr=shg_mgr)

        # Invalidate the data cache.
        self._cache_data_id = None

    def generate_events(
            self,
            rss,
            dataset,
            data,
            mean=None,
            poisson=True,
            tl=None,
    ):
        """Generates a ``mean`` number of background events for the given
        dataset and its data.

        The procedure is as follows:

            1. If the dataset has changed, calculate the background event
               probability for each MC event and cache it.
            1.1. Calculate the mean number of bkg events for the entire MC.
            1.2. Pre-select MC events if a pre event selection method is set.
            1.3. Calculate the background drawing probability of each selected
                 MC event.
            1.4. Calculate the mean number of bkg events for the selected MC
                 events.
            2. Draw the bkg events based on the calculated bkg event
               probability.
            3. Scramble the drawn background events if a data scrambler is set.
            4. Remove any MC specific data fields from the background events
               DataFieldRecordArray instance.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to generate
            random numbers from.
        dataset : instance of Dataset
            The Dataset instance describing the dataset for which background
            events should get generated.
        data : instance of DatasetData
            The DatasetData instance holding the data of the dataset for which
            background events should get generated.
        mean : float | None
            The mean number of background events to generate.
            Can be `None`. In that case the mean number of background events is
            obtained through the `get_mean_func` function.
        poisson : bool
            If set to ``True`` (default), the actual number of generated
            background events will be drawn from a Poisson distribution with the
            given mean number of background events.
            If set to ``False``, the argument ``mean`` specifies the actual
            number of generated background events.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        n_bkg : int
            The number of generated background events for the data set.
        bkg_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the generated
            background events. The number of events can be less than `n_bkg`
            if an event selection method is used.
        """
        tracing = self._cfg['debugging']['enable_tracing']

        # Check if the data set has changed. In that case need to get new
        # background probabilities for each monte-carlo event and a new mean
        # number of background events.
        data_id = id(data)
        if self._cache_data_id != data_id:
            if tracing:
                logger.debug(
                    f'DatasetData instance id of dataset "{dataset.name}" '
                    f'changed from {self._cache_data_id} to {data_id}')
            # Cache the current id of the data.
            self._cache_data_id = data_id

            # Create a copy of the MC data with all MC data fields removed,
            # except the specified MC data fields to keep for the
            # ``get_mean_func`` and ``get_event_prob_func`` functions.
            keep_field_names = list(set(
                DataFields.get_joint_names(
                    datafields=self._cfg['datafields'],
                    stages=(
                        DFS.ANALYSIS_EXP
                    )
                ) +
                data.exp_field_names +
                self._keep_mc_data_field_names
            ))
            self._cache_mc = data.mc.copy(keep_fields=keep_field_names)

            if self._get_mean_func is not None:
                with TaskTimer(
                        tl,
                        'Calculate total MC background mean.'):
                    self._cache_mean = self._get_mean_func(
                        dataset=dataset,
                        data=data,
                        events=self._cache_mc)

            if self._pre_event_selection_method is not None:
                with TaskTimer(
                        tl,
                        'Pre-select MC events.'):
                    (self._cache_mc, _) =\
                        self._pre_event_selection_method.select_events(
                            events=self._cache_mc,
                            ret_original_evt_idxs=False,
                            tl=tl)

                with TaskTimer(tl, 'Calculate selected MC background mean.'):
                    self._cache_mean_pre_selected = self._get_mean_func(
                        dataset=dataset,
                        data=data,
                        events=self._cache_mc)

            with TaskTimer(
                    tl,
                    'Calculate MC background event probability cache.'):
                self._cache_mc_event_bkg_prob = self._get_event_prob_func(
                    dataset=dataset,
                    data=data,
                    events=self._cache_mc)

            with TaskTimer(
                    tl,
                    'Create RandomChoice for MC background events.'):
                self._cache_random_choice = RandomChoice(
                    items=self._cache_mc.indices,
                    probabilities=self._cache_mc_event_bkg_prob)

        if mean is None:
            if self._cache_mean is None:
                raise ValueError(
                    'No mean number of background events and no '
                    'get_mean_func were specified! One of the two must be '
                    'specified!')
            mean = self._cache_mean
        else:
            mean = float_cast(
                mean,
                'The mean number of background events must be cast-able to '
                'type float!')

        # Draw the number of background events from a poisson distribution with
        # the given mean number of background events. This will be the number of
        # background events for this data set.
        n_bkg = (int(rss.random.poisson(mean)) if poisson else
                 int(np.round(mean, 0)))

        # Calculate the mean number of background events for the pre-selected
        # MC events.
        if self._pre_event_selection_method is None:
            # No selection at all, use the total mean.
            mean_pre_selected = mean
        else:
            mean_pre_selected = self._cache_mean_pre_selected

        # Calculate the actual number of background events for the selected
        # events.
        n_bkg_selected = int(np.around(n_bkg * mean_pre_selected / mean, 0))

        # Draw the actual background events from the selected events of the
        # monte-carlo data set.
        with TaskTimer(tl, 'Draw MC background indices.'):
            bkg_event_indices = self._cache_random_choice(
                rss=rss,
                size=n_bkg_selected)

        with TaskTimer(tl, 'Select MC background events from indices.'):
            bkg_events = self._cache_mc[bkg_event_indices]

        # Scramble the drawn MC events if requested.
        if self._data_scrambler is not None:
            with TaskTimer(tl, 'Scramble MC background data.'):
                bkg_events = self._data_scrambler.scramble_data(
                    rss=rss,
                    dataset=dataset,
                    data=bkg_events,
                    copy=False)

        # Remove MC specific data fields from the background events record
        # array. So the result contains only experimental data fields. The list
        # of experimental data fields is defined as the unique set of the
        # required experimental data fields defined by the data set, and the
        # actual experimental data fields (in case there are additional kept
        # data fields by the user).
        with TaskTimer(tl, 'Remove MC specific data fields from MC events.'):
            exp_field_names = list(set(
                DataFields.get_joint_names(
                    datafields=self._cfg['datafields'],
                    stages=(
                        DFS.ANALYSIS_EXP
                    )
                ) +
                data.exp_field_names))
            bkg_events.tidy_up(exp_field_names)

        return (n_bkg, bkg_events)


class CompositeMCDataSamplingBkgGenMethod(
    MCDataSamplingBkgGenMethod,
):
    """This background generation method samples background events from the
    monte-carlo (MC) data. It supports a list of background components, which
    depend on the sky position and hence whose event rates need to be calculated
    for each individual trial after scrambling the MC sample.
    """
    def __init__(
            self,
            bkg_component_rate_calc_func_dict,
            get_event_prob_func,
            get_mean_func=None,
            data_scrambler=None,
            keep_mc_data_fields=None,
            pre_event_selection_method=None,
            **kwargs,
    ):
        """Creates a new instance of CompositeMCDataSamplingBkgGenMethod.

        Parameters
        ----------
        bkg_component_rate_calc_func_dict : dict
            The dictionary holding the name of the background component as key,
            e.g. "gp", and the background rate calculation function for that
            component as value.
            The call signature of each rate calculation function must be

                __call__(dataset, data, events)

            where ``dataset`` is an instance of Dataset and ``data`` is an
            instance of DatasetData of the data set for which background events
            needs to get generated. The ``events`` argument holds the actual
            set of events, for which the background rate needs to get
            calculated.
        get_event_prob_func : callable
            The function to get the background probability of each monte-carlo
            event. The call signature of this function must be

                __call__(dataset, data, events)

            where ``dataset`` is an instance of Dataset and ``data`` is an
            instance of DatasetData of the data set for which background events
            needs to get generated. The ``events`` argument holds the actual
            set of events, for which the background event probabilities need to
            get calculated.
        get_mean_func : callable | None
            The function to get the mean number of background events.
            The call signature of this function must be

                __call__(dataset, data, events)

            where ``dataset`` is an instance of Dataset and ``data`` is an
            instance of DatasetData of the data set for which background events
            needs to get generated. The `events` argument holds the actual set
            of events, for which the mean number of background events should get
            calculated. This argument can be `None`, which means that the mean
            number of background events to generate needs to get specified
            through the ``generate_events`` method. However, if a pre event
            selection method is provided, this argument cannot be ``None``!
        data_scrambler : instance of DataScrambler | None
            If set to an instance of DataScrambler, the monte-carlo events will
            get scrambled before drawing the background events. This can ensure
            more independent data trials. It is especially important when
            monte-carlo statistics are low.
        keep_mc_data_fields : str | list of str | None
            The MC data field names that should be kept in order to be able to
            calculate the background events rates by the functions
            ``get_event_prob_func`` and ``get_mean_func``. All other MC fields
            will get dropped due to computational efficiency reasons.
        pre_event_selection_method : instance of EventSelectionMethod | None
            If set to an instance of EventSelectionMethod, this method will
            pre-select the MC events that will be used for later background
            event generation. Using this pre-selection a large portion of the
            MC data can be reduced prior to background event generation.
        """
        super().__init__(
            get_event_prob_func=get_event_prob_func,
            get_mean_func=get_mean_func,
            data_scrambler=data_scrambler,
            keep_mc_data_fields=keep_mc_data_fields,
            pre_event_selection_method=pre_event_selection_method,
            **kwargs)

        self.bkg_component_rate_calc_func_dict = bkg_component_rate_calc_func_dict

    @property
    def bkg_component_rate_calc_func_dict(self):
        """The dictionary holding the background components (as key) and their
        rate calculation functions (as value).
        """
        return self._bkg_component_rate_calc_func_dict

    @bkg_component_rate_calc_func_dict.setter
    def bkg_component_rate_calc_func_dict(self, d):
        if not isinstance(d, dict):
            raise TypeError(
                'The bkg_component_rate_calc_func_dict property must be an '
                'instance of dict! '
                f'Its current type is {classname(d)}.')
        for (name, func) in d.items():
            if not isinstance(name, str):
                raise TypeError(
                    'The keys of the dictionary of the '
                    'bkg_component_rate_calc_func_dict property must be '
                    'instances of type str! At least one of the keys are of '
                    f'type {classname(name)}!')
            if not func_has_n_args(func, 3):
                raise TypeError(
                    'The function provided for the background component '
                    f'"{name}"  must have 3 arguments!')
        self._bkg_component_rate_calc_func_dict = d

    def generate_events(
            self,
            rss,
            dataset,
            data,
            mean=None,
            poisson=True,
            tl=None,
    ):
        """Generates a ``mean`` number of background events for the given
        monte-carlo dataset and its data.

        The procedure is as follows:

            1. Scramble all MC events if a data scrambler is set.
            2. Calculate the rate for each background component for each MC
               event of the entire MC.
            3. Calculate the mean number of bkg events for the entire MC.
            4. Pre-select MC events if a pre event selection method is set.
            5. Calculate the drawing probability of each selected MC event.
            6. Calculate the mean number of bkg events for the selected MC
               events.
            7. Draw the bkg events based on the calculated bkg event
               probability.
            8. Remove any MC specific data fields.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to generate
            random numbers from.
        dataset : instance of Dataset
            The Dataset instance describing the dataset for which background
            events should get generated.
        data : instance of DatasetData
            The DatasetData instance holding the data of the dataset for which
            background events should get generated.
        mean : float | None
            The mean number of background events to generate.
            Can be `None`. In that case the mean number of background events is
            obtained through the `get_mean_func` function.
        poisson : bool
            If set to ``True`` (default), the actual number of generated
            background events will be drawn from a Poisson distribution with the
            given mean number of background events.
            If set to ``False``, the argument ``mean`` specifies the actual
            number of generated background events.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        n_bkg : int
            The number of generated background events for the data set.
        bkg_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the generated
            background events. The number of events can be less than `n_bkg`
            if an event selection method is used.
        """
        # Create a copy of the MC data with all MC data fields removed,
        # except the specified MC data fields to keep for the
        # ``bkg_component_rate_calc_func``, ``get_mean_func`` and
        # ``get_event_prob_func`` functions.
        keep_field_names = list(set(
            DataFields.get_joint_names(
                datafields=self._cfg['datafields'],
                stages=(
                    DFS.ANALYSIS_EXP
                )
            ) +
            data.exp_field_names +
            self._keep_mc_data_field_names
        ))
        data_mc = data.mc.copy(keep_fields=keep_field_names)

        # Scramble the MC events if requested.
        if self._data_scrambler is not None:
            with TaskTimer(tl, 'Scramble MC events.'):
                data_mc = self._data_scrambler.scramble_data(
                    rss=rss,
                    dataset=dataset,
                    data=data_mc,
                    copy=False)

        # Calculate the rate for each background component.
        # Note: In Python3 the order of the dictionary entries is defined by
        #       the order they were added to the dictionary, hence the order is
        #       preserved.
        for (name, rate_calc_func) in self._bkg_component_rate_calc_func_dict.items():
            rate = rate_calc_func(dataset, data, data_mc)
            data_mc[name] = rate

        # Calculate the mean number of bkg events for the entire MC.
        mean_mc = None
        if self._get_mean_func is not None:
            with TaskTimer(
                    tl,
                    'Calculate total MC background mean.'):
                mean_mc = self._get_mean_func(
                    dataset=dataset,
                    data=data,
                    events=data_mc)

        # Pre-select MC events if a pre event selection method is set.
        if self._pre_event_selection_method is not None:
            with TaskTimer(
                    tl,
                    'Pre-select MC events.'):
                (data_mc, _) =\
                    self._pre_event_selection_method.select_events(
                        events=data_mc,
                        ret_original_evt_idxs=False,
                        tl=tl)

            with TaskTimer(tl, 'Calculate selected MC background mean.'):
                mean_mc_pre_selected = self._get_mean_func(
                    dataset=dataset,
                    data=data,
                    events=data_mc)

        # Calculate the drawing probability of each selected MC event.
        with TaskTimer(
                tl,
                'Calculate MC background event probability.'):
            mc_event_bkg_prob = self._get_event_prob_func(
                dataset=dataset,
                data=data,
                events=data_mc)

        with TaskTimer(
                tl,
                'Create RandomChoice for MC background events.'):
            random_choice = RandomChoice(
                items=data_mc.indices,
                probabilities=mc_event_bkg_prob)

        # Select the correct mean value.
        if mean is None:
            if mean_mc is None:
                raise ValueError(
                    'No mean number of background events and no '
                    'get_mean_func were specified! One of the two must be '
                    'specified!')
            mean = mean_mc
        else:
            mean = float_cast(
                mean,
                'The mean number of background events must be cast-able to '
                'type float!')

        # Draw the number of background events from a poisson distribution with
        # the given mean number of background events. This will be the number of
        # background events for this data set.
        n_bkg = (int(rss.random.poisson(mean)) if poisson else
                 int(np.round(mean, 0)))

        # Calculate the mean number of background events for the pre-selected
        # MC events.
        if self._pre_event_selection_method is None:
            # No selection at all, use the total mean.
            mean_pre_selected = mean
        else:
            mean_pre_selected = mean_mc_pre_selected

        # Calculate the actual number of background events for the selected
        # events.
        n_bkg_selected = int(np.around(n_bkg * mean_pre_selected / mean, 0))

        # Draw the actual background events from the selected events of the
        # monte-carlo data set.
        with TaskTimer(tl, 'Draw MC background indices.'):
            bkg_event_indices = random_choice(
                rss=rss,
                size=n_bkg_selected)

        with TaskTimer(tl, 'Select MC background events from indices.'):
            bkg_events = data_mc[bkg_event_indices]

        # Remove MC specific data fields from the background events record
        # array. So the result contains only experimental data fields. The list
        # of experimental data fields is defined as the unique set of the
        # required experimental data fields defined by the data set, and the
        # actual experimental data fields (in case there are additional kept
        # data fields by the user).
        with TaskTimer(tl, 'Remove MC specific data fields from MC events.'):
            exp_field_names = list(set(
                DataFields.get_joint_names(
                    datafields=self._cfg['datafields'],
                    stages=(
                        DFS.ANALYSIS_EXP
                    )
                ) +
                data.exp_field_names))
            bkg_events.tidy_up(exp_field_names)

        return (n_bkg, bkg_events)



"""
The minimizers/iminuit module provides a SkyLLH interface to the iminuit
minimizer.
"""

import numpy as np

from skyllh.core import (
    tool,
)
from skyllh.core.config import (
    HasConfig,
)
from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core.minimizer import (
    MinimizerImpl,
)


logger = get_logger(__name__)


class FuncWithGradsFunctor(
        HasConfig,
):
    """Helper class to evaluate the LLH function that returns the function value
    and its gradients in two seperate calls, one for the LLH function value and
    one for its gradient values.
    """

    def __init__(
            self,
            func,
            func_args=None,
            **kwargs,
    ):
        """Initializes a new functor instance for the given function ``func``.

        Parameters
        ----------
        func : callable
            The function with call signature

                ``__call__(x, *args)``

            returning the a two-element tuple (f, grads).
        func_args : tuple | None
            The optional positional arguments for the function ``func``.
        """
        super().__init__(
            **kwargs)

        if func_args is None:
            func_args = tuple()

        self._func = func
        self._func_args = func_args

        self._tracing = self._cfg.is_tracing_enabled

        self._cache_x = None
        self._cache_f = None
        self._cache_grads = None

    def get_f(self, x):
        tracing = self._tracing

        if self._cache_x is None:
            self._cache_x = np.copy(x)
        else:
            if np.all(x == self._cache_x):
                if tracing:
                    logger.debug(
                        f'call_func(x={x}): Return cached f={self._cache_f}')
                return self._cache_f
            else:
                np.copyto(self._cache_x, x)

        (self._cache_f, self._cache_grads) = self._func(
            x, *self._func_args)

        if tracing:
            logger.debug(
                f'call_func(x={x}): Return calculated f={self._cache_f}')
        return self._cache_f

    def get_grads(self, x):
        tracing = self._tracing

        if self._cache_x is None:
            self._cache_x = np.copy(x)
        else:
            if np.all(x == self._cache_x):
                if tracing:
                    logger.debug(
                        f'call_grads(x={x}): Return cached '
                        f'grads={self._cache_grads}')
                return self._cache_grads
            else:
                np.copyto(self._cache_x, x)

        (self._cache_f, self._cache_grads) = self._func(
            x, *self._func_args)

        if tracing:
            logger.debug(
                f'call_grads(x={x}): Return calculated '
                f'grads={self._cache_grads}')
        return self._cache_grads


class IMinuitMinimizerImpl(
        MinimizerImpl,
):
    """The SkyLLH minimizer implementation that utilizes the iminuit minimizer.
    """

    @tool.requires('iminuit')
    def __init__(
            self,
            ftol=1e-6,
            **kwargs,
    ):
        """Creates a new IMinuit minimizer instance to minimize a given
        function.

        Parameters
        ----------
        ftol : float
            The function value tolerance as absolute value.
        """
        super().__init__(**kwargs)

        self._ftol = ftol

    def minimize(
            self,
            initials,
            bounds,
            func,
            func_args=None,
            **kwargs,
    ):
        """Minimizes the given function ``func`` with the given initial function
        argument values ``initials`` and within the given parameter bounds
        ``bounds``.

        Parameters
        ----------
        initials : 1D numpy ndarray
            The ndarray holding the initial values of all the fit parameters.
        bounds : 2D (N_fitparams,2)-shaped numpy ndarray
            The ndarray holding the boundary values (vmin, vmax) of the fit
            parameters.
        func : callable
            The function that should get minimized.
            The call signature must be

                ``__call__(x, *args)``

            The return value of ``func`` must be (f, grads), the function value
            at the function arguments ``x`` and the ndarray with the values of
            the function gradient for each fit parameter, if the
            ``func_provides_grads`` keyword argument option is set to True.
            If set to False, ``func`` must return only the function value.
        func_args : sequence | None
            Optional sequence of arguments for ``func``.

        Additional Keyword Arguments
        ----------------------------
        Additional keyword arguments include options for this minimizer
        implementation. Possible options are:

            func_provides_grads : bool
                Flag if the function ``func`` also returns its gradients.
                Default is ``True``.


        Any additional keyword arguments are passed on to the underlaying
        :func:`iminuit.minimize` minimization function.

        Returns
        -------
        xmin : 1D ndarray
            The array containing the function arguments at the function's
            minimum.
        fmin : float
            The function value at its minimum.
        res : iminuit.OptimizeResult
            The iminuit OptimizeResult dictionary with additional information.
        """
        if func_args is None:
            func_args = tuple()
        if kwargs is None:
            kwargs = dict()

        iminuit = tool.get('iminuit')

        func_provides_grads = kwargs.pop('func_provides_grads', True)

        if func_provides_grads:
            # The function func returns the function value and its gradients,
            # so we need to use the FuncWithGradsFunctor helper class.
            functor = FuncWithGradsFunctor(
                cfg=self._cfg,
                func=func,
                func_args=func_args)

            res = iminuit.minimize(
                fun=functor.get_f,
                x0=initials,
                bounds=bounds,
                jac=functor.get_grads,
                tol=self._ftol,
                **kwargs
            )
        else:
            # The function func returns only the function value, so we can use
            # the
            res = iminuit.minimize(
                func,
                initials,
                bounds=bounds,
                args=func_args,
                tol=self._ftol,
                **kwargs
            )

        return (res.x, res.fun, res)

    def get_niter(self, status):
        """Returns the number of iterations needed to find the minimum.

        Parameters
        ----------
        status : dict
            The dictionary with the status information about the minimization
            process.

        Returns
        -------
        niter : int
            The number of iterations needed to find the minimum.
        """
        return status['nfev']

    def has_converged(self, status):
        """Analyzes the status information dictionary if the minimization
        process has converged. By definition the minimization process has
        converged if ``status['is_valid']`` equals True.

        Parameters
        ----------
        status : dict
            The dictionary with the status information about the minimization
            process.

        Returns
        -------
        converged : bool
            The flag if the minimization has converged (True), or not (False).
        """
        return bool(status['success'])

    def is_repeatable(self, status):
        """Checks if the minimization process can be repeated to get a better
        result.

        TODO: Implement a proper check. For now just return True.


        Parameters
        ----------
        status : dict
            The dictionary with the status information about the last
            minimization process.

        Returns
        -------
        repeatable : bool
            The flag if the minimization process can be repeated to obtain a
            better minimum.
        """
        return True



# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.py import (
    classname,
    int_cast,
)


class RandomStateService(
        object):
    """The RandomStateService class provides a container for a
    numpy.random.RandomState object, initialized with a given seed. This service
    can then be passed to any function or method within skyllh that requires a
    random number generator.
    """
    def __init__(
            self,
            seed=None,
            **kwargs,
    ):
        """Creates a new random state service. The ``random`` property can then
        be used to draw random numbers.

        Parameters
        ----------
        seed : int | None
            The seed to use. If None, the random number generator will be seeded
            randomly. See the numpy documentation for numpy.random.RandomState
            what that means.
        """
        super().__init__(**kwargs)

        self._seed = int_cast(
            seed,
            'The seed argument must be None, or cast-able to type int!',
            allow_None=True)
        self.random = np.random.RandomState(self._seed)

    @property
    def seed(self):
        """(read-only) The seed (int) of the random number generator.
        None, if not set. To change the seed, use the `reseed` method.
        """
        return self._seed

    @property
    def random(self):
        """The numpy.random.RandomState object.
        """
        return self._random

    @random.setter
    def random(self, random):
        if not isinstance(random, np.random.RandomState):
            raise TypeError(
                'The random property must be of type numpy.random.RandomState!')
        self._random = random

    def reseed(self, seed):
        """Reseeds the random number generator with the given seed.

        Parameters
        ----------
        seed : int | None
            The seed to use. If None, the random number generator will be seeded
            randomly. See the numpy documentation for numpy.random.RandomState
            what that means.
        """
        self._seed = int_cast(
            seed,
            'The seed argument must be None or cast-able to type int!',
            allow_None=True)
        self.random.seed(self._seed)


class RandomChoice(
        object,
):
    """This class provides an efficient numpy.random.choice functionality
    specialized for SkyLLH. The advantage is that it stores the cumulative
    distribution function (CDF), which is assumed to be constant.
    """

    def __init__(
            self,
            items,
            probabilities,
            **kwargs,
    ):
        """Creates a new instance of RandomChoice holding the probabilities
        and their cumulative distribution function (CDF).

        Parameters
        ----------
        items : instance of numpy.ndarray
            The (N,)-shaped numpy.ndarray holding the items from which to
            choose.
        probabilities : instance of numpy.ndarray
            The (N,)-shaped numpy.ndarray holding the probability for each item.
        """
        super().__init__(**kwargs)

        self._assert_items(items)
        self._items = items

        self._assert_probabilities(probabilities, self._items.size)
        self._probabilities = probabilities

        # Create the cumulative distribution function (CDF). We use float64 to
        # avoid a possible overflow when doing the summation.
        self._cdf = np.cumsum(self._probabilities, dtype=np.float64)
        self._cdf /= self._cdf[-1]

    @property
    def items(self):
        """(read-only) The (N,)-shaped numpy.ndarray holding the items from
        which to choose.
        """
        return self._items

    @property
    def probabilities(self):
        """(read-only) The (N,)-shaped numpy.ndarray holding the probability
        for each item.
        """
        return self._probabilities

    def _assert_items(
            self,
            items,
    ):
        """Checks for the correct type and shape of the items.

        Parameters
        ----------
        items : The (N,)-shaped numpy.ndarray holding the items from which to
            choose.

        Raises
        ------
        TypeError
            If the type of items is not numpy.ndarray.
        ValueError
            If the items do not have the correct type and shape.
        """
        if not isinstance(items, np.ndarray):
            raise TypeError(
                'The items must be an instance of numpy.ndarray! '
                f'Its current type is {classname(items)}!')

        if items.ndim != 1:
            raise ValueError(
                'The items must be a 1-dimensional numpy.ndarray!')

    def _assert_probabilities(
            self,
            p,
            n_items,
    ):
        """Checks for correct values of the probabilities.

        Parameters
        ----------
        p : instance of numpy.ndarray
            The (N,)-shaped numpy.ndarray holding the probability for each item.
        n_items : int
            The number of items.

        Raises
        ------
        ValueError
            If the probabilities do not have the correct type, shape and values.
        """
        atol = np.sqrt(np.finfo(np.float64).eps)
        atol = max(atol, np.sqrt(np.finfo(p.dtype).eps))

        if p.ndim != 1:
            raise ValueError(
                'The probabilities must be provided as a 1-dimensional '
                'numpy.ndarray!')

        if p.size != n_items:
            raise ValueError(
                f'The size ({p.size}) of the probabilities array must match '
                f'the number of items ({n_items})!')

        if np.any(p < 0):
            raise ValueError(
                'The probabilities must be greater or equal zero!')

        p_sum = np.sum(p)
        if abs(p_sum - 1.) > atol:
            raise ValueError(
                f'The sum of the probabilities ({p_sum}) must be 1!')

    def __call__(
            self,
            rss,
            size,
    ):
        """Chooses ``size`` random items from ``self.items`` according to
        ``self.probabilities``.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService from which random numbers are
            drawn from.
        size : int
            The number of items to draw.

        Returns
        -------
        random_items : instance of numpy.ndarray
            The (size,)-shaped numpy.ndarray holding the randomly selected items
            from ``self.items``.
        """
        uniform_values = rss.random.random(size)

        # The np.searchsorted function is much faster when the values are
        # sorted. But we want to keep the randomness of the returned items.
        idxs_of_sort = np.argsort(uniform_values)
        sorted_idxs = np.searchsorted(
            self._cdf,
            uniform_values[idxs_of_sort],
            side='right')
        idxs = np.empty_like(sorted_idxs)
        idxs[idxs_of_sort] = sorted_idxs
        random_items = self._items[idxs]

        return random_items



# -*- coding: utf-8 -*-

"""The trialdata module of SkyLLH provides a trial data manager class that
manages the data of an analysis trial. It provides also possible additional data
fields and their calculation, which are required by the particular analysis.
The rational behind this manager is to compute data fields only once, which can
then be used by different analysis objects, like PDF objects.
"""

from collections import OrderedDict
import numpy as np

from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core import display as dsp
from skyllh.core.py import (
    classname,
    func_has_n_args,
    int_cast,
    issequenceof,
)
from skyllh.core.storage import (
    DataFieldRecordArray,
)


logger = get_logger(__name__)


class DataField(object):
    """This class defines a data field and its calculation that is used by an
    Analysis class instance. The calculation is defined through an external
    function.
    """
    def __init__(
            self,
            name,
            func,
            global_fitparam_names=None,
            dt=None,
            is_src_field=False,
            is_srcevt_data=False,
            **kwargs):
        """Creates a new instance of DataField that might depend on fit
        parameters.

        Parameters
        ----------
        name : str
            The name of the data field. It serves as the identifier for the
            data field.
        func : callable
            The function that calculates the values of this data field. The call
            signature must be

                __call__(tdm, shg_mgr, pmm, global_fitparams_dict=None)

            where ``tdm`` is the instance of TrialDataManager holding the trial
            event data, ``shg_mgr`` is the instance of SourceHypoGroupManager,
            ``pmm`` is the instance of ParameterModelMapper, and
            ``global_fitparams_dict`` is the dictionary with the current global
            fit parameter names and values.
        global_fitparam_names : str | sequence of str | None
            The sequence of str instances specifying the names of the global fit
            parameters this data field depends on. If set to None, the data
            field does not depend on any fit parameters.
        dt : numpy dtype | str | None
            If specified it defines the data type this data field should have.
            If a str instance is given, it defines the name of the data field
            whose data type should be taken for this data field.
        is_src_field : bool
            Flag if this data field is a source data field (``True``) and values
            should be stored within this DataField instance, instead of the
            events DataFieldRecordArray instance of the TrialDataManager
            (``False``).
        is_srcevt_data : bool
            Flag if the data field will hold source-event data, i.e. data of
            length N_values. In that case the data cannot be stored within the
            events attribute of the TrialDataManager, but must be stored in the
            values attribute of this DataField instance.
        """
        super().__init__(**kwargs)

        self.name = name
        self.func = func

        if global_fitparam_names is None:
            global_fitparam_names = []
        if isinstance(global_fitparam_names, str):
            global_fitparam_names = [global_fitparam_names]
        if not issequenceof(global_fitparam_names, str):
            raise TypeError(
                'The global_fitparam_names argument must be None or a sequence '
                'of str instances! It is of type '
                f'{classname(global_fitparam_names)}!')
        self._global_fitparam_name_list = list(global_fitparam_names)

        self.dt = dt

        # Define the list of fit parameter values for which the fit parameter
        # depend data field values have been calculated for.
        self._global_fitparam_value_list = [None] *\
            len(self._global_fitparam_name_list)

        self._is_srcevt_data = is_srcevt_data

        # Define the member variable that holds the numpy ndarray with the data
        # field values.
        self._values = None

        # Define the most efficient `calculate` method for this kind of data
        # field.
        if is_src_field:
            self.calculate = self._calc_source_values
        elif len(self._global_fitparam_name_list) == 0:
            self.calculate = self._calc_static_values
        else:
            self.calculate = self._calc_global_fitparam_dependent_values

    @property
    def name(self):
        """The name of the data field.
        """
        return self._name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError(
                'The name property must be an instance of str!'
                f'It is of type {classname(name)}!')
        self._name = name

    @property
    def func(self):
        """The function that calculates the data field values.
        """
        return self._func

    @func.setter
    def func(self, f):
        if not callable(f):
            raise TypeError(
                'The func property must be a callable object!')
        if (not func_has_n_args(f, 3)) and\
           (not func_has_n_args(f, 4)):
            raise TypeError(
                'The func property must be a function with 3 or 4 arguments!')
        self._func = f

    @property
    def dt(self):
        """The numpy dtype object defining the data type of this data field.
        A str instance defines the name of the data field whose data type should
        be taken for this data field.
        It is None, if there is no explicit data type defined for this data
        field.
        """
        return self._dt

    @dt.setter
    def dt(self, obj):
        if obj is not None:
            if (not isinstance(obj, np.dtype)) and\
               (not isinstance(obj, str)):
                raise TypeError(
                    'The dt property must be None, an instance of numpy.dtype, '
                    'or an instance of str! Currently it is of type '
                    f'{classname(obj)}.')
        self._dt = obj

    @property
    def is_srcevt_data(self):
        """(read-only) Flag if the data field contains source-event data, i.e.
        is of length N_values.
        """
        return self._is_srcevt_data

    @property
    def values(self):
        """(read-only) The calculated data values of the data field.
        """
        return self._values

    def __str__(self):
        """Pretty string representation of this DataField instance.
        """
        dtype = 'None'
        vmin = np.nan
        vmax = np.nan

        if self._values is not None:
            dtype = str(self._values.dtype)
            vmin = np.min(self._values)
            vmax = np.max(self._values)

        s = f'{classname(self)}: {self.name}: '
        s += '{dtype: '
        s += f'{dtype}, vmin: {vmin: .3e}, vmax: {vmax: .3e}'
        s += '}'

        return s

    def _get_desired_dtype(self, tdm):
        """Retrieves the data type this field should have. It's ``None``, if no
        data type was defined for this data field.
        """
        if self._dt is not None:
            if isinstance(self._dt, str):
                # The _dt attribute defines the name of the data field whose
                # data type should be used.
                self._dt = tdm.get_dtype(self._dt)
        return self._dt

    def _convert_to_desired_dtype(self, tdm, values):
        """Converts the data type of the given values array to the given data
        type.
        """
        dt = self._get_desired_dtype(tdm)
        if dt is not None:
            values = values.astype(dt, copy=False)
        return values

    def _calc_source_values(
            self,
            tdm,
            shg_mgr,
            pmm):
        """Calculates the data field values utilizing the defined external
        function. The data field values solely depend on fixed source
        parameters.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance this data field is part of and is
            holding the event data.
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the source
            hypothesis groups.
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper, which defines the global
            parameters and their mapping to local source parameters.
        """
        self._values = self._func(
            tdm=tdm,
            shg_mgr=shg_mgr,
            pmm=pmm)

        if not isinstance(self._values, np.ndarray):
            raise TypeError(
                f'The calculation function for the data field "{self._name}" '
                'must return an instance of numpy.ndarray! '
                f'Currently it is of type "{classname(self._values)}".')

        # Convert the data type.
        self._values = self._convert_to_desired_dtype(tdm, self._values)

    def _calc_static_values(
            self,
            tdm,
            shg_mgr,
            pmm):
        """Calculates the data field values utilizing the defined external
        function, that are static and only depend on source parameters.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance this data field is part of and is
            holding the event data.
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the source
            hypothesis groups.
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper, which defines the global
            parameters and their mapping to local source parameters.
        """
        values = self._func(
            tdm=tdm,
            shg_mgr=shg_mgr,
            pmm=pmm)

        if not isinstance(values, np.ndarray):
            raise TypeError(
                f'The calculation function for the data field "{self._name}" '
                'must return an instance of numpy.ndarray! '
                f'Currently it is of type "{classname(values)}".')

        # Convert the data type.
        values = self._convert_to_desired_dtype(tdm, values)

        if self._is_srcevt_data:
            n_values = tdm.get_n_values()
            if values.shape[0] != n_values:
                raise ValueError(
                    'The calculation function for the data field '
                    f'"{self._name}" must return a numpy ndarray of shape '
                    f'({n_values},), but the shape is {values.shape}!')
            self._values = values
        else:
            # Set the data values. This will add the data field to the
            # DataFieldRecordArray if it does not exist yet.
            tdm.events[self._name] = values

    def _calc_global_fitparam_dependent_values(
            self,
            tdm,
            shg_mgr,
            pmm,
            global_fitparams_dict):
        """Calculate data field values utilizing the defined external
        function, that depend on fit parameter values. We check if the fit
        parameter values have changed.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance this data field is part of and is
            holding the event data.
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the source
            hypothesis groups.
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper defining the mapping of the
            global parameters to local source parameters.
        global_fitparams_dict : dict
            The dictionary holding the current global fit parameter names and
            values.
        """
        # Determine if we need to calculate the values.
        calc_values = False

        if self._name not in tdm.events:
            calc_values = True
        else:
            for (idx, name) in enumerate(self._global_fitparam_name_list):
                if global_fitparams_dict[name] !=\
                   self._global_fitparam_value_list[idx]:
                    calc_values = True
                    break

        if not calc_values:
            return

        values = self._func(
            tdm=tdm,
            shg_mgr=shg_mgr,
            pmm=pmm,
            global_fitparams_dict=global_fitparams_dict)

        if not isinstance(values, np.ndarray):
            raise TypeError(
                'The calculation function for the data field '
                f'"{self._name}" must return an instance of numpy.ndarray! '
                f'Currently it is of type "{classname(values)}".')

        # Convert the data type.
        values = self._convert_to_desired_dtype(tdm, values)

        if self._is_srcevt_data:
            n_values = tdm.get_n_values()
            if values.shape[0] != n_values:
                raise ValueError(
                    'The calculation function for the data field '
                    f'"{self._name}" must return a numpy ndarray of shape '
                    f'({n_values},), but the shape is {values.shape}!')
            self._values = values
        else:
            # Set the data values. This will add the data field to the
            # DataFieldRecordArray if it does not exist yet.
            tdm.events[self._name] = values

        # We store the global fit parameter values for which the field values
        # were calculated. So they have to get recalculated only when the
        # global fit parameter values, the field depends on, change.
        self._global_fitparam_value_list = [
            global_fitparams_dict[name]
            for name in self._global_fitparam_name_list
        ]


class TrialDataManager(object):
    """The TrialDataManager class manages the event data for an analysis trial.
    It provides possible additional data fields and their calculation.
    New data fields can be defined via the :py:meth:`add_data_field` method.
    Whenever a new trial is being initialized the data fields get re-calculated.
    The data trial manager is provided to the PDF evaluation method.
    Hence, data fields are calculated only once.
    """
    def __init__(self, index_field_name=None, **kwargs):
        """Creates a new TrialDataManager instance.

        Parameters
        ----------
        index_field_name : str | None
            The name of the field that should be used as primary index field.
            If provided, the events will be sorted along this data field. This
            might be useful for run-time performance.
        """
        super().__init__(**kwargs)

        self.index_field_name = index_field_name

        # Define the list of data fields that depend only on the source
        # parameters.
        self._source_data_fields_dict = OrderedDict()

        # Define the list of data fields that are static and should be
        # calculated prior to a possible event selection.
        self._pre_evt_sel_static_data_fields_dict = OrderedDict()

        # Define the list of data fields that are static, i.e. don't depend on
        # any fit parameters. These fields have to be calculated only once when
        # a new evaluation data is available.
        self._static_data_fields_dict = OrderedDict()

        # Define the list of data fields that depend on global fit parameters.
        # These data fields have to be re-calculated whenever a global fit
        # parameter value changes.
        self._global_fitparam_data_fields_dict = OrderedDict()

        # Define the member variable that will hold the number of sources.
        self._n_sources = None

        # Define the member variable that will hold the total number of events
        # of the dataset this TrialDataManager belongs to.
        self._n_events = None

        # Define the member variable that will hold the raw events for which the
        # data fields get calculated.
        self._events = None

        # Define the member variable that holds the source to event index
        # mapping.
        self._src_evt_idxs = None

        # We store an integer number for the trial data state and increase it
        # whenever the state of the trial data changed. This way other code,
        # e.g. PDFs, can determine when the data changed and internal caches
        # must be flushed.
        self._trial_data_state_id = -1

    @property
    def index_field_name(self):
        """The name of the primary index data field. If not None, events will
        be sorted by this data field.
        """
        return self._index_field_name

    @index_field_name.setter
    def index_field_name(self, name):
        if name is not None:
            if not isinstance(name, str):
                raise TypeError(
                    'The index_field_name property must be an instance of '
                    f'type str! It is of type {classname(name)}!')
        self._index_field_name = name

    @property
    def events(self):
        """The DataFieldRecordArray instance holding the data events, which
        should get evaluated.
        """
        return self._events

    @events.setter
    def events(self, arr):
        if not isinstance(arr, DataFieldRecordArray):
            raise TypeError(
                'The events property must be an instance of '
                f'DataFieldRecordArray! It is of type {classname(arr)}!')
        self._events = arr

    @property
    def has_global_fitparam_data_fields(self):
        """(read-only) ``True`` if the TrialDataManager has global fit parameter
        data fields defined, ``False`` otherwise.
        """
        return len(self._global_fitparam_data_fields_dict) > 0

    @property
    def n_sources(self):
        """(read-only) The number of sources. This information is taken from
        the source hypo group manager when a new trial is initialized.
        """
        return self._n_sources

    @property
    def n_events(self):
        """The total number of events of the dataset this trial data manager
        corresponds to.
        """
        return self._n_events

    @n_events.setter
    def n_events(self, n):
        self._n_events = int_cast(
            n, 'The n_events property must be castable to type int!')

    @property
    def n_selected_events(self):
        """(read-only) The number of selected events which should get evaluated.
        """
        return len(self._events)

    @property
    def n_pure_bkg_events(self):
        """(read-only) The number of pure background events, which are not part
        of the trial data, but must be considered for the test-statistic value.
        It is the difference of n_events and n_selected_events.
        """
        return self._n_events - len(self._events)

    @property
    def src_evt_idxs(self):
        """(read-only) The 2-tuple holding the source indices and event indices
        1d ndarray arrays. This can be ``None``, indicating that all trial data
        events should be considered for all sources.
        """
        return self._src_evt_idxs

    @property
    def trial_data_state_id(self):
        """(read-only) The integer ID number of the trial data. This ID number
        can be used to determine when the trial data has changed its state.
        """
        return self._trial_data_state_id

    def __contains__(self, name):
        """Checks if the given data field is defined in this data field manager.

        Parameters
        ----------
        name : str
            The name of the data field.

        Returns
        -------
        check : bool
            True if the data field is defined in this data field manager,
            False otherwise.
        """
        # Check if the data field is part of the original trial data.
        if (self._events is not None) and\
           (name in self._events.field_name_list):
            return True

        # Check if the data field is a user defined data field.
        if (name in self._source_data_fields_dict) or\
           (name in self._pre_evt_sel_static_data_fields_dict) or\
           (name in self._static_data_fields_dict) or\
           (name in self._global_fitparam_data_fields_dict):
            return True

        return False

    def __getitem__(self, name):
        """Implements the evaluation of ``self[name]`` to access data fields.
        This method calls the :meth:`get_data` method of this class.
        """
        return self.get_data(name)

    def __str__(self):
        """Implements pretty string representation of this TrialDataManager
        instance.
        """
        s = classname(self)+':\n'
        s1 = 'Base data fields:\n'
        s2 = str(self._events)
        s1 += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s2)
        s += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s1)
        s += '\n'

        s1 = 'Source data fields:\n'
        s2 = '\n'.join(
            [
                str(df)
                for (_, df) in self._source_data_fields_dict.items()
            ]
        )
        if s2 == '':
            s2 = 'None'
        s1 += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s2)
        s += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s1)
        s += '\n'

        s1 = 'Pre-event-selection static data fields:\n'
        s2 = '\n'.join(
            [
                str(df)
                for (_, df) in self._pre_evt_sel_static_data_fields_dict.items()
            ]
        )
        if s2 == '':
            s2 = 'None'
        s1 += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s2)
        s += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s1)
        s += '\n'

        s1 = 'Static data fields:\n'
        s2 = '\n'.join(
            [
                str(df)
                for (_, df) in self._static_data_fields_dict.items()
            ]
        )
        if s2 == '':
            s2 = 'None'
        s1 += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s2)
        s += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s1)
        s += '\n'

        s1 = 'Global fitparam data fields:\n'
        s2 = '\n'.join(
            [
                str(df)
                for (_, df) in self._global_fitparam_data_fields_dict.items()
            ]
        )
        if s2 == '':
            s2 = 'None'
        s1 += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s2)
        s += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s1)

        return s

    def broadcast_sources_array_to_values_array(
            self,
            arr):
        """Broadcasts the given 1d numpy ndarray of length 1 or N_sources to a
        numpy ndarray of length N_values.

        Parameters
        ----------
        arr : instance of ndarray
            The (N_sources,)- or (1,)-shaped numpy ndarray holding values for
            each source.

        Returns
        -------
        out_arr : instance of ndarray
            The (N_values,)-shaped numpy ndarray holding the source values
            broadcasted to each event value.
        """
        arr_dtype = arr.dtype
        n_values = self.get_n_values()

        if len(arr) == 1:
            return np.full((n_values,), arr[0], dtype=arr_dtype)

        if len(arr) != self.n_sources:
            raise ValueError(
                f'The length of arr ({len(arr)}) must be 1 or equal to the '
                f'number of sources ({self.n_sources})!')

        out_arr = np.empty(
            (n_values,),
            dtype=arr.dtype)

        src_idxs = self.src_evt_idxs[0]
        v_start = 0
        for (src_idx, src_value) in enumerate(arr):
            n = np.count_nonzero(src_idxs == src_idx)
            # n = len(evt_idxs[src_idxs == src_idx])
            out_arr[v_start:v_start+n] = np.full(
                (n,), src_value, dtype=arr_dtype)
            v_start += n

        return out_arr

    def broadcast_sources_arrays_to_values_arrays(
            self,
            arrays):
        """Broadcasts the 1d numpy ndarrays to the values array.

        Parameters
        ----------
        arrays : sequence of numpy 1d ndarrays
            The sequence of (N_sources,)-shaped numpy ndarrays holding the
            parameter values.

        Returns
        -------
        out_arrays : list of numpy 1d ndarrays
            The list of (N_values,)-shaped numpy ndarrays holding the
            broadcasted array values.
        """
        out_arrays = [
            self.broadcast_sources_array_to_values_array(arr)
            for arr in arrays
        ]

        return out_arrays

    def broadcast_selected_events_arrays_to_values_arrays(
            self,
            arrays):
        """Broadcasts the given arrays of length N_selected_events to arrays
        of length N_values.

        Parameters
        ----------
        arrays : sequence of instance of ndarray
            The sequence of instance of ndarray with the arrays to be
            broadcasted.

        Returns
        -------
        out_arrays : list of instance of ndarray
            The list of broadcasted numpy ndarray instances.
        """
        evt_idxs = self._src_evt_idxs[1]
        out_arrays = [
            np.take(arr, evt_idxs)
            for arr in arrays
        ]

        return out_arrays

    def change_shg_mgr(self, shg_mgr, pmm):
        """This method is called when the source hypothesis group manager has
        changed. Hence, the source data fields need to get recalculated.

        After calling this method, a new trial should be initialized via the
        :meth:`initialize_trial` method!

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the source
            hypothesis groups.
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper that defines the global
            parameters and their mapping to local source parameter.
        """
        self.calculate_source_data_fields(
            shg_mgr=shg_mgr,
            pmm=pmm)

    def initialize_trial(
            self,
            shg_mgr,
            pmm,
            events,
            n_events=None,
            evt_sel_method=None,
            tl=None):
        """Initializes the trial data manager for a new trial. It sets the raw
        events, calculates pre-event-selection data fields, performs a possible
        event selection and calculates the static data fields for the left-over
        events.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the source
            hypothesis groups.
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper, that defines the global
            parameters and their mapping to local source parameters.
        events : DataFieldRecordArray instance
            The DataFieldRecordArray instance holding the entire raw events.
        n_events : int | None
            The total number of events of the data set this trial data manager
            corresponds to.
            If None, the number of events is taken from the number of events
            present in the ``events`` array.
        evt_sel_method : instance of EventSelectionMethod | None
            The optional event selection method that should be used to select
            potential signal events.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used for timing
            measurements.
        """
        # Set the events property, so that the calculation functions of the data
        # fields can access them.
        self.events = events
        self._src_evt_idxs = None

        # Save the number of sources.
        self._n_sources = shg_mgr.n_sources

        if n_events is None:
            n_events = len(self._events)
        self.n_events = n_events

        # Calculate pre-event-selection data fields that are required by the
        # event selection method.
        self.calculate_pre_evt_sel_static_data_fields(
            shg_mgr=shg_mgr,
            pmm=pmm)

        if evt_sel_method is not None:
            logger.debug(
                f'Performing event selection method '
                f'"{classname(evt_sel_method)}".')
            (selected_events, src_evt_idxs) = evt_sel_method.select_events(
                events=self._events,
                tl=tl)
            logger.debug(
                f'Selected {len(selected_events)} out of {len(self._events)} '
                'events.')
            self.events = selected_events
            self._src_evt_idxs = src_evt_idxs

        # Sort the events by the index field, if a field was provided.
        if self._index_field_name is not None:
            logger.debug(
                f'Sorting events in index field "{self._index_field_name}"')
            sorted_idxs = self._events.sort_by_field(self._index_field_name)
            # If event indices are stored, we need to re-assign also those event
            # indices according to the new order.
            if self._src_evt_idxs is not None:
                self._src_evt_idxs[1] = np.take(
                    sorted_idxs, self._src_evt_idxs[1])

        # Create the src_evt_idxs property data in case it was not provided by
        # the event selection. In that case all events are selected for all
        # sources. This simplifies the implementations of the PDFs.
        if self._src_evt_idxs is None:
            self._src_evt_idxs = (
                np.repeat(np.arange(self.n_sources), self.n_selected_events),
                np.tile(np.arange(self.n_selected_events), self.n_sources)
            )

        # Now calculate all the static data fields. This will increment the
        # trial data state ID.
        self.calculate_static_data_fields(
            shg_mgr=shg_mgr,
            pmm=pmm)

    def get_n_values(self):
        """Returns the expected size of the values array after a PDF
        evaluation, which will include PDF values for all trial data events and
        all sources.

        Returns
        -------
        n : int
            The length of the expected values array after a PDF evaluation.
        """
        return len(self._src_evt_idxs[0])

    def get_values_mask_for_source_mask(self, src_mask):
        """Creates a boolean mask for the values array where entries belonging
        to the sources given by the source mask are selected.

        Parameters
        ----------
        src_mask : instance of numpy ndarray
            The (N_sources,)-shaped numpy ndarray holding the boolean selection
            of the sources.

        Returns
        -------
        values_mask : instance of numpy ndarray
            The (N_values,)-shaped numpy ndarray holding the boolean selection
            of the values.
        """
        tdm_src_idxs = self.src_evt_idxs[0]
        src_idxs = np.arange(self.n_sources)[src_mask]

        values_mask = np.zeros((self.get_n_values(),), dtype=np.bool_)

        def make_values_mask(src_idx):
            global values_mask
            values_mask |= tdm_src_idxs == src_idx

        np.vectorize(make_values_mask)(src_idxs)

        return values_mask

    def add_source_data_field(
            self,
            name,
            func,
            dt=None):
        """Adds a new data field to the manager. The data field must depend
        solely on source parameters.

        Parameters
        ----------
        name : str
            The name of the data field. It serves as the identifier for the
            data field.
        func : callable
            The function that calculates the data field values. The call
            signature must be

                __call__(tdm, shg_mgr, pmm)

            where ``tdm`` is the TrialDataManager instance holding the event
            data, ``shg_mgr`` is the instance of SourceHypoGroupManager,
            and ``pmm`` is the instance of ParameterModelMapper.
        dt : numpy dtype | str | None
            If specified it defines the data type this data field should have.
            If a str instance is given, it defines the name of the data field
            whose data type should be taken for the data field.
        """
        if name in self:
            raise KeyError(
                f'The data field "{name}" is already defined!')

        data_field = DataField(
            name=name,
            func=func,
            dt=dt,
            is_src_field=True)

        self._source_data_fields_dict[name] = data_field

    def add_data_field(
            self,
            name,
            func,
            global_fitparam_names=None,
            dt=None,
            pre_evt_sel=False,
            is_srcevt_data=False):
        """Adds a new data field to the manager.

        Parameters
        ----------
        name : str
            The name of the data field. It serves as the identifier for the
            data field.
        func : callable
            The function that calculates the data field values. The call
            signature must be

                __call__(tdm, shg_mgr, pmm, global_fitparams_dict=None)

            where ``tdm`` is the TrialDataManager instance holding the trial
            event data, ``shg_mgr`` is the instance of SourceHypoGroupManager,
            ``pmm`` is the instance of ParameterModelMapper, and
            ``global_fitparams_dict`` is the dictionary with the current global
            fit parameter names and values.
            The shape of the returned array must be (N_selected_events,).
        global_fitparam_names : str | sequence of str | None
            The sequence of str instances specifying the names of the global fit
            parameters this data field depends on. If set to ``None``, it means
            that the data field does not depend on any fit parameters.
        dt : numpy dtype | str | None
            If specified it defines the data type this data field should have.
            If a str instance is given, it defines the name of the data field
            whose data type should be taken for the data field.
        pre_evt_sel : bool
            Flag if this data field should get calculated before potential
            signal events get selected (True), or afterwards (False).
            Default is False.
        is_srcevt_data : bool
            Flag if this data field contains source-event data, hence the length
            of the data array will be N_values.
            Default is False.
        """
        if name in self:
            raise KeyError(
                'The data field "{name}" is already defined!')

        if pre_evt_sel:
            if global_fitparam_names is not None:
                raise ValueError(
                    f'The pre-event-selection data field "{name}" must not '
                    'depend on global fit parameters!')

            if is_srcevt_data:
                raise ValueError(
                    'By definition the pre-event-selection data field '
                    f'"{name}" cannot hold source-event data! The '
                    'is_srcevt_data argument must be set to False!')

        data_field = DataField(
            name=name,
            func=func,
            global_fitparam_names=global_fitparam_names,
            dt=dt,
            is_src_field=False,
            is_srcevt_data=is_srcevt_data)

        if pre_evt_sel:
            self._pre_evt_sel_static_data_fields_dict[name] = data_field
        elif global_fitparam_names is None:
            self._static_data_fields_dict[name] = data_field
        else:
            self._global_fitparam_data_fields_dict[name] = data_field

    def calculate_source_data_fields(
            self,
            shg_mgr,
            pmm):
        """Calculates the data values of the data fields that solely depend on
        source parameters.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses.
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper, that defines the global
            parameters and their mapping to local source parameters.
        """
        if len(self._source_data_fields_dict) == 0:
            return

        for (name, dfield) in self._source_data_fields_dict.items():
            dfield.calculate(
                tdm=self,
                shg_mgr=shg_mgr,
                pmm=pmm)

        self._trial_data_state_id += 1

    def calculate_pre_evt_sel_static_data_fields(
            self,
            shg_mgr,
            pmm):
        """Calculates the data values of the data fields that should be
        available for the event selection method and do not depend on any fit
        parameters.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses.
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper, that defines the global
            parameters and their mapping to local source parameters.
        """
        if len(self._pre_evt_sel_static_data_fields_dict) == 0:
            return

        for (name, dfield) in self._pre_evt_sel_static_data_fields_dict.items():
            dfield.calculate(
                tdm=self,
                shg_mgr=shg_mgr,
                pmm=pmm)

        self._trial_data_state_id += 1

    def calculate_static_data_fields(
            self,
            shg_mgr,
            pmm):
        """Calculates the data values of the data fields that do not depend on
        any source or fit parameters.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses.
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper, that defines the global
            parameters and their mapping to local source parameters.
        """
        if len(self._static_data_fields_dict) == 0:
            return

        for (name, dfield) in self._static_data_fields_dict.items():
            dfield.calculate(
                tdm=self,
                shg_mgr=shg_mgr,
                pmm=pmm)

        self._trial_data_state_id += 1

    def calculate_global_fitparam_data_fields(
            self,
            shg_mgr,
            pmm,
            global_fitparams_dict):
        """Calculates the data values of the data fields that depend on global
        fit parameter values.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses.
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper, that defines the global
            parameters and their mapping to local source parameters.
        global_fitparams_dict : dict
            The dictionary holding the current global fit parameter names and
            values.
        """
        if len(self._global_fitparam_data_fields_dict) == 0:
            return

        for (name, dfield) in self._global_fitparam_data_fields_dict.items():
            dfield.calculate(
                tdm=self,
                shg_mgr=shg_mgr,
                pmm=pmm,
                global_fitparams_dict=global_fitparams_dict)

        self._trial_data_state_id += 1

    def get_data(self, name):
        """Gets the data for the given data field name. The data is stored
        either in the raw events DataFieldRecordArray or in one of the
        additional defined data fields. Data from the raw events
        DataFieldRecordArray is prefered.

        Parameters
        ----------
        name : str
            The name of the data field for which to retrieve the data.

        Returns
        -------
        data : instance of numpy ndarray
            The numpy ndarray holding the data of the requested data field.
            The length of the array is either N_sources, N_selected_events, or
            N_values.

        Raises
        ------
        KeyError
            If the given data field is not defined.
        """
        # Data fields which are static or depend on global fit parameters are
        # stored within the _events DataFieldRecordArray if they do not contain
        # source-event data. For all other cases, the data is stored in the
        # .values attribute of the DataField class instance.
        if self._events is not None and\
           name in self._events.field_name_list:
            return self._events[name]

        if name in self._source_data_fields_dict:
            return self._source_data_fields_dict[name].values

        if name in self._static_data_fields_dict:
            return self._static_data_fields_dict[name].values

        if name in self._global_fitparam_data_fields_dict:
            return self._global_fitparam_data_fields_dict[name].values

        raise KeyError(
            f'The data field "{name}" is not defined!')

    def get_dtype(self, name):
        """Gets the data type of the given data field.

        Parameters
        ----------
        name : str
            The name of the data field whose data type should get retrieved.

        Returns
        -------
        dt : numpy dtype
            The numpy dtype object of the given data field.

        Raises
        ------
        KeyError
            If the given data field is not defined.
        """
        dt = self.get_data(name).dtype

        return dt

    def is_event_data_field(self, name):
        """Checks if the given data field is an events data field, i.e. its
        length is N_selected_events.

        Parameters
        ----------
        name : str
            The name of the data field.

        Returns
        -------
        check : bool
            ``True`` if the given data field contains event data, ``False``
            otherwise.
        """
        if self._events is not None and\
           name in self._events.field_name_list:
            return True

        return False

    def is_source_data_field(self, name):
        """Checks if the given data field is a source data field, i.e. its
        length is N_sources.

        Parameters
        ----------
        name : str
            The name of the data field.

        Returns
        -------
        check : bool
            ``True`` if the given data field contains source data, ``False``
            otherwise.
        """
        if name in self._source_data_fields_dict:
            return True

        return False

    def is_srcevt_data_field(self, name):
        """Checks if the given data field is a source-event data field, i.e. its
        length is N_values.

        Parameters
        ----------
        name : str
            The name of the data field.

        Returns
        -------
        check : bool
            ``True`` if the given data field contains source-event data,
            ``False`` otherwise.
        """
        if name in self._static_data_fields_dict:
            return self._static_data_fields_dict[name].is_srcevt_data

        if name in self._global_fitparam_data_fields_dict:
            return self._global_fitparam_data_fields_dict[name].is_srcevt_data

        return False



# -*- coding: utf-8 -*-

"""This module provides functionality for defining which data fields of a data
file is required at what stage.
"""


class DataFieldStages(
    object,
):
    """This class provides the data field stage values, which are individual
    bits of an integer value to allow for multiple stages of a data field.
    """

    DATAPREPARATION_EXP = 1
    DATAPREPARATION_MC = 2
    ANALYSIS_EXP = 4
    ANALYSIS_MC = 8

    @staticmethod
    def and_check(
            stage,
            stages,
    ):
        """Checks if the given stage matches all of the given stages.

        Parameters
        ----------
        stage : int
            The stage value, which should get checked.
        stages : int | sequence of int
            The stage(s) to check for.

        Returns
        -------
        check : bool
            ``True`` if the given stage contains all of the given stages,
            ``False`` otherwise.
        """
        if isinstance(stages, int):
            return (stage & stages == stages)

        for stage_ in stages:
            if stage & stage_ != stage_:
                return False

        return True

    @staticmethod
    def or_check(
            stage,
            stages,
    ):
        """Checks if the given stage matches any of the given stages.

        Parameters
        ----------
        stage : int
            The stage value, which should get checked.
        stages : int | sequence of int
            The stage(s) to check for.

        Returns
        -------
        check : bool
            ``True`` if the given stage contains any of the given stages,
            ``False`` otherwise.
        """
        if isinstance(stages, int):
            return (stage & stages != 0)

        for stage_ in stages:
            if stage & stage_ != 0:
                return True

        return False


class DataFields(
    object,
):
    @staticmethod
    def get_joint_names(
            datafields,
            stages,
    ):
        """Returns the list of data field names that match at least one of the
        given stages, i.e. the joint set of data fields given the stages.

        Parameters
        ----------
        datafields : dict
            The dictionary of data field names as keys and stages as values.
        stages : int | sequence of int
            The stage(s) for which data field names should get returned.

        Returns
        -------
        datafield_names : list of str
            The list of data field names.
        """
        datafield_names = [
            field
            for (field, stage) in datafields.items()
            if DataFieldStages.or_check(stage, stages)
        ]

        return datafield_names



# -*- coding: utf-8 -*-

import abc

from skyllh.core.py import (
    issequence,
    float_cast,
)


class SignalGenerationMethod(
        object,
        metaclass=abc.ABCMeta
):
    """This is a base class for a source and detector specific signal generation
    method, that calculates the source flux for a given monte-carlo event, which
    is needed to calculate the MC event weights for the signal generator.
    """

    def __init__(
            self,
            energy_range,
            **kwargs,
    ):
        """Constructs a new signal generation method instance.

        Parameters
        ----------
        energy_range : 2-element tuple of float | None
            The energy range from which to take MC events into account for
            signal event generation.
            If set to None, the entire energy range [0, +inf] is used.
        """
        super().__init__(**kwargs)

        self.energy_range = energy_range

    @property
    def energy_range(self):
        """The 2-element tuple of floats holding the energy range from which to
        take MC events into account for signal event generation.
        """
        return self._energy_range

    @energy_range.setter
    def energy_range(self, r):
        if r is not None:
            if not issequence(r):
                raise TypeError(
                    'The energy_range property must be a sequence!')
            if len(r) != 2:
                raise ValueError(
                    'The energy_range property must be a sequence of 2 '
                    'elements!')
            r = (
                float_cast(
                    r[0],
                    'The first element of the energy_range '
                    'sequence must be castable to type float!'),
                float_cast(
                    r[1],
                    'The second element of the energy_range '
                    'sequence must be castable to type float!')
            )
        self._energy_range = r

    @abc.abstractmethod
    def calc_source_signal_mc_event_flux(
            self,
            data_mc,
            shg,
    ):
        """This method is supposed to calculate the signal flux of each given
        MC event for each source hypothesis of the given source hypothesis
        group.

        Parameters
        ----------
        data_mc : numpy record ndarray
            The numpy record array holding all the MC events.
        shg : instance of SourceHypoGroup
            The source hypothesis group instance, which defines the list of
            sources, and their flux model.

        Returns
        -------
        ev_idx_arr : ndarray
            The (N_selected_signal_events,)-shaped 1D ndarray holding the index
            of the MC event.
        shg_src_idx_arr : ndarray
            The (N_selected_signal_events,)-shaped 1D ndarray holding the index
            of the source within the given source hypothesis group for each
            signal candidate event.
        flux_arr : ndarray
            The (N_selected_signal_events,)-shaped 1D ndarray holding the flux
            value of each signal candidate event.
        """
        pass

    def signal_event_post_sampling_processing(
        self,
        shg,
        shg_sig_events_meta,
        shg_sig_events,
    ):
        """This method should be reimplemented by the derived class if there
        is some processing needed after the MC signal events have been sampled
        from the global MC data.

        Parameters
        ----------
        shg : SourceHypoGroup instance
            The source hypothesis group instance holding the sources and their
            locations.
        shg_sig_events_meta : numpy record ndarray
            The numpy record ndarray holding meta information about the
            generated signal events for the given source hypothesis group.
            The length of this array must be the same as shg_sig_events.
            It needs to contain the following data fields:

                shg_src_idx : int
                    The source index within the source hypothesis group.

        shg_sig_events : numpy record ndarray
            The numpy record ndarray holding the generated signal events for
            the given source hypothesis group and in the format of the original
            MC events.

        Returns
        -------
        shg_sig_events : numpy record array
            The processed signal events. In the default implementation of this
            method this is just the shg_sig_events input array.
        """
        return shg_sig_events



# -*- coding: utf-8 -*-

from tqdm import tqdm

from skyllh.core import (
    session,
)
from skyllh.core.py import (
    int_cast,
)


class ProgressBar(
        object):
    """This class provides an hierarchical progress bar for SkyLLH.
    For rendering it uses the tqdm Python package.
    In case of multiple layers of progress bars, it creates only a single
    progress bar, which gets updated whenever the deeper level progress bars
    are updated.
    """
    def __init__(
            self,
            maxval,
            startval=0,
            parent=None,
            **kwargs):
        """Creates a new ProgressBar instance.

        Parameters
        ----------
        maxval : int
            The maximal value the progress can reach.
        startval : int
            The progress value to start with. Must be smaller than `maxval`.
        parent : instance of ProgressBar | False | None
            The parent instance of ProgressBar if this progress bar is a sub
            progress bar.
            If set to ``False``, this progress bar is deactivated and the
            property ``is_shown`` will return ``False``.
            If set to ``None``, this progress bar will be a primary progress
            bar.

        Additional keyword arguments
        ----------------------------
        Additional keyword arguments are passed to the constructor of the tqdm
        class.
        """
        super().__init__(
            **kwargs)

        self._is_deactivated = False
        if parent is False:
            self._is_deactivated = True
            parent = None

        self.maxval = maxval
        self.startval = startval
        self.parent = parent

        self._val = 0
        self._sub_pbar_list = []

        self._tqdm = None
        if (self._parent is None) and self.is_shown:
            self._tqdm = tqdm(
                total=maxval,
                initial=startval,
                leave=True,
                position=0,
                **kwargs)

    @property
    def maxval(self):
        """The maximal integer value the progress can reach.
        """
        return self._maxval

    @maxval.setter
    def maxval(self, v):
        v = int_cast(
            v,
            'The maxval property must be castable to an integer value!')
        self._maxval = v

    @property
    def startval(self):
        """The progress integer value to start with. It must be smaller than
        `maxval`.
        """
        return self._startval

    @startval.setter
    def startval(self, v):
        v = int_cast(
            v,
            'The startval property must be castable to an integer value!')

        if v >= self._maxval:
            raise ValueError(
                f'The startval value ({v}) must be smaller than the value of '
                f'the `maxval` property ({self._maxval})!')

        self._startval = v

    @property
    def val(self):
        """(read-only) The current value of the progess.
        """
        return self._val

    @property
    def is_shown(self):
        """(read-only) Flag if the progress bar is shown. This is ``True``
        if the program is run in an interactive session, ``False`` otherwise.
        """
        if self._is_deactivated is True:
            return False
        return session.is_interactive_session()

    @property
    def parent(self):
        """The parent ProgressBar instance of this progress bar, or ``None`` if
        no parent exist.
        """
        return self._parent

    @parent.setter
    def parent(self, pbar):
        if pbar is not None:
            if not isinstance(pbar, ProgressBar):
                raise TypeError(
                    'The parent property must be None, or an instance of '
                    'ProgressBar!')
        self._parent = pbar

    def add_sub_progress_bar(self, pbar):
        """Adds the given progress bar to the list of running sub progress bars
        of this progress bar.
        """
        if not isinstance(pbar, ProgressBar):
            raise TypeError(
                'The pbar argument must be an instance of ProgressBar!')
        self._sub_pbar_list.append(pbar)

    def remove_sub_progress_bars(self):
        """Removes all progress bar instances from the list of sub progress bars
        of this progress bar. It calles the ``remove_sub_progress_bars`` method
        of each sub progress bar.
        """
        for pbar in self._sub_pbar_list:
            pbar.remove_sub_progress_bars()

        self._sub_pbar_list = []

    def get_progressbar_list(self):
        """Retrieves the list of ProgressBar instances.

        Returns
        -------
        pbar_list : list of instance of ProgressBar
            The list of ProgressBar instances, which are part of this
            ProgressBar instance.
        """
        pbar_list = []
        for pbar in self._sub_pbar_list:
            pbar_list.extend(pbar.get_progressbar_list())

        pbar_list.append(self)

        return pbar_list

    def rerender(self):
        """Rerenders this progress bar on the display. It calls the ``update``
        method of the tqdm progess bar.
        """
        if not self.is_shown:
            return

        pbar_list = self.get_progressbar_list()

        maxval = 0
        val = 0
        for pbar in pbar_list:
            maxval += pbar.maxval
            val += pbar.val

        dval = val - self._tqdm.n
        self._tqdm.total = maxval
        self._tqdm.update(dval)

    def trigger_rerendering(self):
        """Triggers a rerendering / update of the most top progress bar.
        """
        if self._parent is not None:
            self._parent.trigger_rerendering()
            return

        # We are the most top parent progress bar. So we need to get rerendered.
        self.rerender()

    def start(self):
        """Sets the current progess value to ``startval`` and updates the
        progess bar to fulfill the start conditions.
        """
        self._val = self._startval

        if self._parent is not None:
            self._parent.add_sub_progress_bar(self)
        elif not self.is_shown:
            return self
        else:
            self._tqdm.initial = self._val
            self._tqdm.n = self._val
            self._tqdm.reset()

        self.trigger_rerendering()

        return self

    def finish(self):
        """Finishes this progress bar by setting the current progress value to
        its max value.
        If this progress bar is the top most progessbar, it will also close
        the tqdm instance, what will trigger a flush of the output buffer.
        """
        self._val = self._maxval

        self.trigger_rerendering()

        if (self._parent is None) and self.is_shown:
            self._tqdm.close()

        self.remove_sub_progress_bars()

    def increment(self, dval=1):
        """Updates the progress bar by incrementing the progress by the given
        integral amount.

        Parameters
        ----------
        dval : int
            The amount of progress to increment the progress bar with.
        """
        self.update(self._val + dval)

    def update(self, val):
        """Updates the progress value to the given value.

        Parameters
        ----------
        val : int
            The new current progress value.
        """
        self._val = val

        self.trigger_rerendering()



# -*- coding: utf-8 -*-

"""The display module provides global settings for pretty command line
displaying.
"""

# Define the width (number of characters) of the display.
PAGE_WIDTH = 80

# Define the width (number of characters) for each text block indentation.
INDENTATION_WIDTH = 4


class ANSIColors:
    """This class defines the ANSI color codes, which can be used to change
    the text color in a terminal.
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def add_leading_text_line_padding(padwidth, text):
    """Adds leading white spaces to all the lines of the given text.

    Parameters
    ----------
    padwidth : int
        The width of the padding.
    text : str
        The text with new line characters for each line.

    Returns
    -------
    padded_text : str
        The text where each line is padded with the given number of whitespaces.
    """
    return '\n'.join([' '*padwidth + line for line in text.split('\n')])



# -*- coding: utf-8 -*-

import pickle

from skyllh.core.py import (
    str_cast,
)


class Message(object):
    @staticmethod
    def receive(sock, blocksize=2048, as_bytes=False):
        """Receives a message from the given socket.

        Parameters
        ----------
        blocksize : int
            The size in bytes of the block that should be read from the socket
            at once.
        as_bytes : bool
            If set to ``True`` the Message instance will contain a bytes
            message, otherwise it is converted to a str.

        Returns
        -------
        m : Message
            The Message instance created with the message read from the socket.
        """
        # Get the first 2 bytes to determine the length of the message.
        msglen = int.from_bytes(
            read_from_socket(sock, 2, blocksize=blocksize), 'little')

        # Read the message of length msglen bytes from the socket. Here, msg is
        # a bytes object.
        msg = read_from_socket(sock, msglen, blocksize=blocksize)

        if as_bytes:
            return Message(msg)

        return Message(str(msg, 'utf-8'))

    def __init__(self, msg):
        """Creates a new Message instance.

        Parameters
        ----------
        msg : str | bytes
            The message string of this Message instance.
        """
        self.msg = msg

    @property
    def msg(self):
        """The message string. This is either a bytes instance or a str
        instance.
        """
        return self._msg

    @msg.setter
    def msg(self, m):
        if not isinstance(m, bytes):
            m = str_cast(
                m,
                'The msg property must be of type bytes or castable to type '
                'str!')
        self._msg = m

    @property
    def length(self):
        """The length of the message in bytes.
        """
        return len(self.msg)

    def as_socket_msg(self):
        """Converts this message to a bytes instance that can be send through a
        socket. The first two bytes hold the length of the message.
        """
        smsg = len(self.msg).to_bytes(2, 'little')
        if isinstance(self.msg, bytes):
            smsg += self.msg
        else:
            smsg += bytes(self.msg, 'utf-8')

        return smsg

    def send(self, sock):
        send_to_socket(sock, self.as_socket_msg())


def send_to_socket(sock, msg):
    msglen = len(msg)
    n_bytes_sent = 0
    while n_bytes_sent < msglen:
        sent = sock.send(msg[n_bytes_sent:])
        if sent == 0:
            raise RuntimeError('Socket connection broken!')
        n_bytes_sent += sent


def read_from_socket(sock, size, blocksize=2048):
    """Reads ``size`` bytes from the socket ``sock``.
    """
    chunks = []
    n_bytes_recd = 0
    while (n_bytes_recd < size):
        chunk = sock.recv(min(size - n_bytes_recd, blocksize))
        if chunk == b'':
            raise RuntimeError('Socket connection broken!')
        chunks.append(chunk)
        n_bytes_recd += len(chunk)
    return b''.join(chunks)


def receive_object_from_socket(sock, blocksize=2048):
    """Receives a pickled Python object from the given socket.

    Parameters
    ----------
    sock : socket
    """
    m = Message.receive(sock, blocksize, as_bytes=True)
    obj = pickle.loads(m.msg)
    return obj



# -*- coding: utf-8 -*-

"""Plotting module to plot IceCube specific PDF ratio objects.
"""

import numpy as np
import itertools

from matplotlib.axes import Axes
from matplotlib.colors import LogNorm

from skyllh.core.py import classname
from skyllh.core.source_hypo_grouping import (
    SourceHypoGroupManager,
)
from skyllh.core.storage import DataFieldRecordArray
from skyllh.core.trialdata import TrialDataManager
from skyllh.i3.pdfratio import SplinedI3EnergySigSetOverBkgPDFRatio


class SplinedI3EnergySigSetOverBkgPDFRatioPlotter(object):
    """Plotter class to plot an I3EnergySigSetOverBkgPDFRatioSpline object.
    """
    def __init__(self, tdm, pdfratio):
        """Creates a new plotter object for plotting an
        I3EnergySigSetOverBkgPDFRatioSpline object.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that provides the data for the
            PDF ratio evaluation.
        pdfratio : I3EnergySigSetOverBkgPDFRatioSpline
            The PDF ratio object to plot.
        """
        self.tdm = tdm
        self.pdfratio = pdfratio

    @property
    def pdfratio(self):
        """The PDF ratio object to plot.
        """
        return self._pdfratio

    @pdfratio.setter
    def pdfratio(self, pdfratio):
        if not isinstance(pdfratio, SplinedI3EnergySigSetOverBkgPDFRatio):
            raise TypeError(
                'The pdfratio property must be an instance of '
                'SplinedI3EnergySigSetOverBkgPDFRatio!')
        self._pdfratio = pdfratio

    @property
    def tdm(self):
        """The TrialDataManager that provides the data for the PDF evaluation.
        """
        return self._tdm

    @tdm.setter
    def tdm(self, obj):
        if not isinstance(obj, TrialDataManager):
            raise TypeError(
                'The tdm property must be an instance of TrialDataManager!')
        self._tdm = obj

    def plot(self, src_hypo_group_manager, axes, fitparams, **kwargs):
        """Plots the PDF ratio for the given set of fit paramater values.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the source
            hypotheses.
        axes : mpl.axes.Axes
            The matplotlib Axes object on which the PDF ratio should get drawn
            to.
        fitparams : dict
            The dictionary with the set of fit paramater values.

        Additional Keyword Arguments
        ----------------------------
        Any additional keyword arguments will be passed to the `mpl.imshow`
        function.

        Returns
        -------
        img : instance of mpl.AxesImage
            The AxesImage instance showing the PDF ratio image.
        """
        if not isinstance(src_hypo_group_manager, SourceHypoGroupManager):
            raise TypeError(
                'The src_hypo_group_manager argument must be an '
                'instance of SourceHypoGroupManager!')
        if not isinstance(axes, Axes):
            raise TypeError(
                'The axes argument must be an instance of '
                'matplotlib.axes.Axes!')
        if not isinstance(fitparams, dict):
            raise TypeError(
                'The fitparams argument must be an instance of dict!')

        # Get the binning for the axes. We use the background PDF to get it
        # from. By construction, all PDFs use the same binning. We know that
        # the PDFs are 2-dimensional.
        (xbinning, ybinning) = self._pdfratio.backgroundpdf.binnings

        # Create a 2D array with the ratio values. We put one event into each
        # bin.
        ratios = np.zeros((xbinning.nbins, ybinning.nbins), dtype=np.float64)
        events = DataFieldRecordArray(np.zeros(
            (ratios.size,),
            dtype=[('ix', np.int64), (xbinning.name, np.float64),
                   ('iy', np.int64), (ybinning.name, np.float64)]))
        for (i, ((ix, x), (iy, y))) in enumerate(itertools.product(
                enumerate(xbinning.bincenters),
                enumerate(ybinning.bincenters))):
            events['ix'][i] = ix
            events[xbinning.name][i] = x
            events['iy'][i] = iy
            events[ybinning.name][i] = y

        self._tdm.initialize_for_new_trial(src_hypo_group_manager, events)

        event_ratios = self.pdfratio.get_ratio(self._tdm, fitparams)
        for i in range(len(events)):
            ratios[events['ix'][i], events['iy'][i]] = event_ratios[i]

        (left, right, bottom, top) = (xbinning.lower_edge, xbinning.upper_edge,
                                      ybinning.lower_edge, ybinning.upper_edge)
        img = axes.imshow(
            ratios.T,
            extent=(left, right, bottom, top),
            origin='lower',
            norm=LogNorm(),
            interpolation='none',
            **kwargs)
        axes.set_xlabel(xbinning.name)
        axes.set_ylabel(ybinning.name)
        axes.set_title(classname(self._pdfratio))

        return img



# -*- coding: utf-8 -*-

import scipy.interpolate

import numpy as np

from numpy.lib.recfunctions import (
    repack_fields,
)

from skyllh.core.multiproc import (
    IsParallelizable,
    parallelize,
)
from skyllh.core.pdfratio import (
    SigSetOverBkgPDFRatio,
)
from skyllh.core.pdfratio_fill import (
    MostSignalLikePDFRatioFillMethod,
    PDFRatioFillMethod,
)
from skyllh.core.py import (
    make_dict_hash,
)


class SplinedI3EnergySigSetOverBkgPDFRatio(
        SigSetOverBkgPDFRatio,
        IsParallelizable):
    """This class implements a splined signal over background PDF ratio for
    enegry PDFs of type I3EnergyPDF.
    It takes an instance, which is derived from PDFSet, and which is derived
    from IsSignalPDF, as signal PDF. Furthermore, it takes an instance, which
    is derived from I3EnergyPDF and IsBackgroundPDF, as background PDF, and
    creates a spline for the ratio of the signal and background PDFs for a grid
    of different discrete energy signal parameters, which are defined by the
    signal PDF set.
    """
    def __init__(
            self,
            sig_pdf_set,
            bkg_pdf,
            fillmethod=None,
            interpolmethod_cls=None,
            ncpu=None,
            ppbar=None,
            **kwargs,
    ):
        """Creates a new IceCube signal-over-background energy PDF ratio spline
        instance.

        Parameters
        ----------
        sig_pdf_set : class instance derived from PDFSet (for PDF type
                       I3EnergyPDF), IsSignalPDF, and UsesBinning
            The PDF set, which provides signal energy PDFs for a set of
            discrete signal parameters.
        bkg_pdf : class instance derived from I3EnergyPDF, and
                        IsBackgroundPDF
            The background energy PDF object.
        fillmethod : instance of PDFRatioFillMethod | None
            An instance of class derived from PDFRatioFillMethod that implements
            the desired ratio fill method.
            If set to None (default), the default ratio fill method
            MostSignalLikePDFRatioFillMethod will be used.
        interpolmethod_cls : class of GridManifoldInterpolationMethod
            The class implementing the parameter interpolation method for
            the PDF ratio manifold grid.
        ncpu : int | None
            The number of CPUs to use to create the ratio splines for the
            different sets of signal parameters.
        ppbar : ProgressBar instance | None
            The instance of ProgressBar of the optional parent progress bar.

        Raises
        ------
        ValueError
            If the signal and background PDFs use different binning.
        """
        super().__init__(
            sig_pdf_set=sig_pdf_set,
            bkg_pdf=bkg_pdf,
            interpolmethod_cls=interpolmethod_cls,
            ncpu=ncpu,
            **kwargs)

        # Define the default ratio fill method.
        if fillmethod is None:
            fillmethod = MostSignalLikePDFRatioFillMethod()
        self.fillmethod = fillmethod

        # Ensure same binning of signal and background PDFs.
        for sig_pdf in self._sig_pdf_set.values():
            if not sig_pdf.has_same_binning_as(self._bkg_pdf):
                raise ValueError(
                    'At least one signal PDF does not have the same binning '
                    'as the background PDF!')

        def create_log_ratio_spline(
                sig_pdf_set,
                bkg_pdf,
                fillmethod,
                gridparams):
            """Creates the signal/background ratio spline for the given signal
            parameters.

            Returns
            -------
            log_ratio_spline : instance of RegularGridInterpolator
                The spline of the logarithmic PDF ratio values.
            """
            # Get the signal PDF for the given signal parameters.
            sig_pdf = sig_pdf_set[gridparams]

            # Create the ratio array with the same shape than the background pdf
            # histogram.
            ratio = np.ones_like(bkg_pdf.hist, dtype=np.float64)

            # Fill the ratio array.
            ratio = fillmethod(
                ratio,
                sig_pdf.hist,
                bkg_pdf.hist,
                sig_pdf.hist_mask_mc_covered,
                sig_pdf.hist_mask_mc_covered_zero_physics,
                bkg_pdf.hist_mask_mc_covered,
                bkg_pdf.hist_mask_mc_covered_zero_physics)

            # Define the grid points for the spline. In general, we use the bin
            # centers of the binning, but for the first and last point of each
            # dimension we use the lower and upper bin edge, respectively, to
            # ensure full coverage of the spline across the binning range.
            points_list = []
            for binning in sig_pdf.binnings:
                points = binning.bincenters
                (points[0], points[-1]) = (
                    binning.lower_edge, binning.upper_edge)
                points_list.append(points)

            # Create the spline for the ratio values.
            log_ratio_spline = scipy.interpolate.RegularGridInterpolator(
                tuple(points_list),
                np.log(ratio),
                method='linear',
                bounds_error=False,
                fill_value=0.)

            return log_ratio_spline

        # Get the list of parameter permutations on the grid for which we
        # need to create PDF ratio splines.
        gridparams_list = self._sig_pdf_set.gridparams_list

        args_list = [
            ((self._sig_pdf_set,
              self._bkg_pdf,
              self._fillmethod,
              gridparams),
             {})
            for gridparams in gridparams_list
        ]

        log_ratio_spline_list = parallelize(
            func=create_log_ratio_spline,
            args_list=args_list,
            ncpu=self.ncpu,
            ppbar=ppbar)

        # Save all the log_ratio splines in a dictionary.
        self._gridparams_hash_log_ratio_spline_dict = dict()
        for (gridparams, log_ratio_spline) in zip(gridparams_list,
                                                  log_ratio_spline_list):
            gridparams_hash = make_dict_hash(gridparams)
            self._gridparams_hash_log_ratio_spline_dict[gridparams_hash] =\
                log_ratio_spline

        # Save the list of data field names.
        self._data_field_names = [
            binning.name
            for binning in self._bkg_pdf.binnings
        ]

        # Construct the instance for the parameter interpolation method.
        self._interpolmethod = self._interpolmethod_cls(
            func=self._evaluate_splines,
            param_grid_set=sig_pdf_set.param_grid_set)

        # Save the parameter names needed for the interpolation for later usage.
        self._interpol_param_names = \
            self._sig_pdf_set.param_grid_set.params_name_list

        # Create cache variable for the last ratio values and gradients in order
        # to avoid the recalculation of the ratio value when the
        # ``get_gradient`` method is called (usually after the ``get_ratio``
        # method was called).
        self._cache = self._create_cache(
            trial_data_state_id=None,
            interpol_params_recarray=None,
            ratio=None,
            grads=None
        )

    @property
    def fillmethod(self):
        """The PDFRatioFillMethod object, which should be used for filling the
        PDF ratio bins.
        """
        return self._fillmethod

    @fillmethod.setter
    def fillmethod(self, obj):
        if not isinstance(obj, PDFRatioFillMethod):
            raise TypeError(
                'The fillmethod property must be an instance of '
                'PDFRatioFillMethod!')
        self._fillmethod = obj

    def _create_cache(
            self,
            trial_data_state_id,
            interpol_params_recarray,
            ratio,
            grads):
        """Creates a cache dictionary holding cache data.

        Parameters
        ----------
        trial_data_state_id : int | None
            The trial data state ID of the TrialDataManager.
        interpol_params_recarray : instance of numpy record ndarray | None
            The numpy record ndarray of length N_sources holding the parameter
            names and values necessary for the interpolation for all sources.
        ratio : instance of numpy ndarray
            The (N_values,)-shaped numpy ndarray holding the PDF ratio values
            for all sources and trial events.
        grads : instance of numpy ndarray
            The (D,N_values)-shaped numpy ndarray holding the gradients for each
            PDF ratio value w.r.t. each interpolation parameter.
        """
        cache = {
            'trial_data_state_id': trial_data_state_id,
            'interpol_params_recarray': interpol_params_recarray,
            'ratio': ratio,
            'grads': grads
        }

        return cache

    def _is_cached(self, trial_data_state_id, interpol_params_recarray):
        """Checks if the ratio and gradients for the given set of interpolation
        parameters are already cached.
        """
        if self._cache['trial_data_state_id'] is None:
            return False

        if self._cache['trial_data_state_id'] != trial_data_state_id:
            return False

        if not np.all(
                self._cache['interpol_params_recarray'] ==
                interpol_params_recarray):
            return False

        return True

    def _get_spline_for_param_values(self, interpol_param_values):
        """Retrieves the spline for a given set of parameter values.

        Parameters
        ----------
        interpol_param_values : instance of numpy ndarray
            The (N_interpol_params,)-shaped numpy ndarray holding the values of
            the interpolation parameters.

        Returns
        -------
        spline : instance of scipy.interpolate.RegularGridInterpolator
            The requested spline instance.
        """
        gridparams = dict(
            zip(self._interpol_param_names, interpol_param_values))
        gridparams_hash = make_dict_hash(gridparams)

        spline = self._gridparams_hash_log_ratio_spline_dict[gridparams_hash]

        return spline

    def _evaluate_splines(
            self,
            tdm,
            eventdata,
            gridparams_recarray,
            n_values):
        """For each set of parameter values given by ``gridparams_recarray``,
        the spline is retrieved and evaluated for the events suitable for that
        source model.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data and the event
            mapping to the sources via the ``src_evt_idx`` property.
        eventdata : instance of numpy ndarray
            The (V,N_events)-shaped numpy ndarray holding the event data, where
            N_events is the number of events, and V the dimensionality of the
            event data.
        gridparams_recarray : instance of numpy structured ndarray
            The numpy structured ndarray of length N_sources with the parameter
            names and values needed for the interpolation on the grid for all
            sources. If the length of this record array is 1, the set of
            parameters will be used for all sources.
        n_values : int
            The size of the output array.

        Returns
        -------
        values : instance of ndarray
            The (N_values,)-shaped numpy ndarray holding the values for each set
            of parameter values of the ``gridparams_recarray``. The length of
            the array depends on the ``src_evt_idx`` property of the
            TrialDataManager. In the worst case it is
            ``N_sources * N_selected_events``.
        """
        (src_idxs, evt_idxs) = tdm.src_evt_idxs

        # Check for special case when a single set of parameters are provided.
        if len(gridparams_recarray) == 1:
            # We got a single parameter set. We will use it for all sources.
            spline = self._get_spline_for_param_values(gridparams_recarray[0])

            eventdata = np.take(eventdata, evt_idxs, axis=1)
            values = spline(eventdata.T)

            return values

        values = np.empty(n_values, dtype=np.float64)

        v_start = 0
        for (sidx, param_values) in enumerate(gridparams_recarray):
            spline = self._get_spline_for_param_values(param_values)

            # Select the eventdata that belongs to the current source.
            m = src_idxs == sidx
            src_eventdata = np.take(eventdata, evt_idxs[m], axis=1)

            n = src_eventdata.shape[1]
            sl = slice(v_start, v_start+n)
            values[sl] = spline(src_eventdata.T)

            v_start += n

        return values

    def _create_interpol_params_recarray(self, src_params_recarray):
        """Creates the params_recarray needed for the interpolation. It selects
        The interpolation parameters from the ``params_recarray`` argument.
        If all parameters have the same value for all sources, the length will
        be 1.

        Parameters
        ----------
        src_params_recarray : instance of numpy record ndarray
            The numpy record ndarray of length N_sources holding all local
            parameter names and values.

        Returns
        -------
        interpol_params_recarray : instance of numpy record ndarray
            The numpy record ndarray of length N_sources or 1 holding only the
            parameters needed for the interpolation.
        """
        interpol_params_recarray = repack_fields(
            src_params_recarray[self._interpol_param_names])

        all_params_are_equal_for_all_sources = True
        for pname in self._interpol_param_names:
            if not np.all(
                    np.isclose(np.diff(interpol_params_recarray[pname]), 0)):
                all_params_are_equal_for_all_sources = False
                break
        if all_params_are_equal_for_all_sources:
            return interpol_params_recarray[:1]

        return interpol_params_recarray

    def _calculate_ratio_and_grads(
            self,
            tdm,
            interpol_params_recarray):
        """Calculates the ratio values and ratio gradients for all the sources
        and trial events given the source parameter values.
        The result is stored in the class member variable ``_cache``.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data.
        interpol_params_recarray : instance of numpy record ndarray
            The numpy record ndarray of length N_sources holding the parameter
            names and values for all sources.
            It must contain only the parameters necessary for the interpolation.
        """
        # Create a 2D event data array holding only the needed event data fields
        # for the PDF ratio spline evaluation.
        eventdata = np.vstack([tdm[fn] for fn in self._data_field_names])

        (ratio, grads) = self._interpolmethod(
            tdm=tdm,
            eventdata=eventdata,
            params_recarray=interpol_params_recarray)

        # The interpolation works on the logarithm of the ratio spline, hence
        # we need to transform it using the exp function, and we need to account
        # for the exp function in the gradients.
        ratio = np.exp(ratio)
        grads = ratio * grads

        # Cache the value and the gradients.
        self._cache = self._create_cache(
            trial_data_state_id=tdm.trial_data_state_id,
            interpol_params_recarray=interpol_params_recarray,
            ratio=ratio,
            grads=grads
        )

    def get_ratio(
            self,
            tdm,
            src_params_recarray,
            tl=None):
        """Retrieves the PDF ratio values for each given trial event data, given
        the given set of fit parameters. This method is called during the
        likelihood maximization process.
        For computational efficiency reasons, the gradients are calculated as
        well and will be cached.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial event data for which
            the PDF ratio values should get calculated.
        src_params_recarray : instance of numpy record ndarray | None
            The (N_sources,)-shaped numpy record ndarray holding the parameter
            names and values of the sources. See the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            for more information.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        ratio : instance of numpy ndarray
            The (N_values,)-shaped numpy ndarray of float holding the PDF ratio
            value for each source and trial event.
        """
        # Select only the parameters necessary for the interpolation.
        interpol_params_recarray = self._create_interpol_params_recarray(
            src_params_recarray)

        # Check if the ratio values are already cached.
        if self._is_cached(
               trial_data_state_id=tdm.trial_data_state_id,
               interpol_params_recarray=interpol_params_recarray):
            return self._cache['ratio']

        self._calculate_ratio_and_grads(
            tdm=tdm,
            interpol_params_recarray=interpol_params_recarray)

        return self._cache['ratio']

    def get_gradient(
            self,
            tdm,
            src_params_recarray,
            fitparam_id,
            tl=None):
        """Retrieves the PDF ratio gradient for the given fit parameter
        ``fitparam_id``.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial event data for which
            the PDF ratio gradient values should get calculated.
        src_params_recarray : instance of numpy record ndarray | None
            The (N_sources,)-shaped numpy record ndarray holding the local
            parameter names and values of all sources. See the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information.
        fitparam_id : int
            The ID of the global fit parameter for which the gradient should
            get calculated.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        grad : instance of ndarray
            The (N_values,)-shaped numpy ndarray holding the gradient values
            for all sources and trial events w.r.t. the given global fit
            parameter.
        """
        # Select only the parameters necessary for the interpolation.
        interpol_params_recarray = self._create_interpol_params_recarray(
            src_params_recarray)

        # Calculate the gradients if necessary.
        if not self._is_cached(
            trial_data_state_id=tdm.trial_data_state_id,
            interpol_params_recarray=interpol_params_recarray
        ):
            self._calculate_ratio_and_grads(
                tdm=tdm,
                interpol_params_recarray=interpol_params_recarray)

        tdm_n_sources = tdm.n_sources

        grad = np.zeros((tdm.get_n_values(),), dtype=np.float64)

        # Loop through the parameters of the signal PDF set and match them with
        # the global fit parameter.
        for (pidx, pname) in enumerate(
                self._sig_pdf_set.param_grid_set.params_name_list):
            if pname not in src_params_recarray.dtype.fields:
                continue
            p_gpidxs = src_params_recarray[f'{pname}:gpidx']
            src_mask = p_gpidxs == (fitparam_id + 1)
            n_sources = np.count_nonzero(src_mask)
            if n_sources == 0:
                continue
            if n_sources == tdm_n_sources:
                # This parameter applies to all sources, hence to all values,
                # and hence it's the only local parameter contributing to the
                # global parameter fitparam_id.
                return self._cache['grads'][pidx]

            # The current parameter does not apply to all sources.
            # Create a values mask that matches a given source mask.
            values_mask = tdm.get_values_mask_for_source_mask(src_mask)
            grad[values_mask] = self._cache['grads'][pidx][values_mask]

        return grad



# -*- coding: utf-8 -*-

"""This module contains classes for source hypothesis grouping functionalities.
Same kind sources can be grouped to allow more efficient calculations in the
analysis.
"""

import numpy as np

from skyllh.core.display import (
    add_leading_text_line_padding,
    INDENTATION_WIDTH,
)
from skyllh.core.py import (
    classname,
    issequenceof,
)
from skyllh.core.detsigyield import (
    DetSigYieldBuilder,
)
from skyllh.core.flux_model import (
    FluxModel,
)
from skyllh.core.signal_generation import (
    SignalGenerationMethod,
)
from skyllh.core.source_model import (
    SourceModel,
)
from skyllh.core.types import (
    SourceHypoGroup_t,
)


class SourceHypoGroup(
        SourceHypoGroup_t):
    """The source hypothesis group class provides a data container to describe
    a group of sources that share the same flux model, detector signal yield,
    and signal generation methods.
    """

    def __init__(
            self,
            sources,
            fluxmodel,
            detsigyield_builders,
            sig_gen_method=None,
            **kwargs):
        """Constructs a new source hypothesis group.

        Parameters
        ----------
        sources : SourceModel | sequence of SourceModel
            The source or sequence of sources that define the source group.
        fluxmodel : instance of FluxModel
            The FluxModel instance that applies to the list of sources of the
            group.
        detsigyield_builders : sequence of DetSigYieldBuilder instances
            The sequence of detector signal yield builder instances,
            which should be used to create the detector signal
            yield for the sources of this group. Each element is the
            detector signal yield builder for the particular dataset, if
            several datasets are used. If this list contains only one builder,
            it should be used for all datasets.
        sig_gen_method : SignalGenerationMethod instance | None
            The instance of SignalGenerationMethod that implements the signal
            generation for the specific detector and source hypothesis. It can
            be set to None, which means, no signal can be generated. Useful for
            data unblinding and data trial generation, where no signal is
            required.
        """
        self.source_list = sources
        self.fluxmodel = fluxmodel
        self.detsigyield_builder_list = detsigyield_builders
        self.sig_gen_method = sig_gen_method

    @property
    def source_list(self):
        """The list of SourceModel instances for which the group is defined.
        """
        return self._source_list

    @source_list.setter
    def source_list(self, sources):
        if isinstance(sources, SourceModel):
            sources = [sources]
        if not issequenceof(sources, SourceModel):
            raise TypeError(
                'The source_list property must be an instance of SourceModel '
                'or a sequence of SourceModel instances! '
                f'Its current type is {classname(sources)}.')
        self._source_list = list(sources)

    @property
    def fluxmodel(self):
        """The FluxModel instance that applies to the list of sources of this
        source group.
        """
        return self._fluxmodel

    @fluxmodel.setter
    def fluxmodel(self, fluxmodel):
        if not isinstance(fluxmodel, FluxModel):
            raise TypeError(
                'The fluxmodel property must be an instance of FluxModel! '
                f'Its current type is {classname(fluxmodel)}.')
        self._fluxmodel = fluxmodel

    @property
    def detsigyield_builder_list(self):
        """The list of DetSigYieldBuilder instances, which should be used to
        create the detector signal yield for this group of sources. Each
        element is the detector signal yield builder for the particular dataset,
        if several datasets are used. If this list contains only one builder,
        it should be used for all datasets.
        """
        return self._detsigyield_builder_list

    @detsigyield_builder_list.setter
    def detsigyield_builder_list(self, builders):
        if isinstance(builders, DetSigYieldBuilder):
            builders = [builders]
        if not issequenceof(builders, DetSigYieldBuilder):
            raise TypeError(
                'The detsigyield_builder_list property must be a sequence of '
                'DetSigYieldBuilder instances!')
        self._detsigyield_builder_list = builders

    @property
    def sig_gen_method(self):
        """The instance of SignalGenerationMethod that implements the signal
        generation for the specific detector and source hypothesis. It can
        be None, which means, no signal can be generated. Useful for
        data unblinding and data trial generation, where no signal is
        required.
        """
        return self._sig_gen_method

    @sig_gen_method.setter
    def sig_gen_method(self, method):
        if method is not None:
            if not isinstance(method, SignalGenerationMethod):
                raise TypeError(
                    'The sig_gen_method property must be an instance of '
                    'SignalGenerationMethod! '
                    f'Its current type is {classname(method)}.')
        self._sig_gen_method = method

    @property
    def n_sources(self):
        """(read-only) The number of sources within this source hypothesis
        group.
        """
        return len(self._source_list)

    def __str__(self):
        """Pretty string representation of this SourceHypoGroup instance.
        """
        s = f'{classname(self)}:\n'

        s1 = f'sources ({len(self._source_list)}):'
        for (idx, source) in enumerate(self._source_list):
            s1 += '\n'
            s2 = f'{idx}: {source}'
            s1 += add_leading_text_line_padding(INDENTATION_WIDTH, s2)
        s1 += '\n'
        s1 += 'fluxmodel:\n'
        s2 = f'{self._fluxmodel}'
        s1 += add_leading_text_line_padding(INDENTATION_WIDTH, s2)
        s1 += '\n'
        s1 += f'detector signal yield builders ({len(self._detsigyield_builder_list)}):\n'
        s2 = '\n'.join((classname(builder) for builder in self._detsigyield_builder_list))
        s1 += add_leading_text_line_padding(INDENTATION_WIDTH, s2)
        s1 += '\n'
        s1 += 'signal generation method:\n'
        s2 = f'{classname(self._sig_gen_method)}'
        s1 += add_leading_text_line_padding(INDENTATION_WIDTH, s2)

        s += add_leading_text_line_padding(INDENTATION_WIDTH, s1)

        return s

    def get_source_weights(self):
        """Gets the weight from each source of this source hypothesis group.

        Returns
        -------
        weights : numpy ndarray | None
            The (N_sources,)-shaped numpy ndarray holding the theoretical
            weight of each source.
            It is ``None`` if any of the individual source weights is None.
        """
        weights = []
        for src in self._source_list:
            if src.weight is None:
                return None
            weights.append(src.weight)

        return np.array(weights)


class SourceHypoGroupManager(
        object):
    """The source hypothesis group manager provides the functionality to group
    sources of the same source hypothesis, i.e. spatial model and flux model,
    with an assigned detector signal yield implementation method.

    This helps to evaluate the log-likelihood ratio function in an efficient
    way.
    """
    def __init__(
            self,
            src_hypo_groups=None,
            **kwargs):
        """Creates a new source hypothesis group manager instance.

        Parameters
        ----------
        src_hypo_groups : SourceHypoGroup instance |
                          sequence of SourceHypoGroup instances | None
            The SourceHypoGroup instances to initialize the manager with.
        """
        super().__init__(**kwargs)

        self._shg_list = list()
        # Define a 2D numpy array of shape (N_sources,2) that maps the source
        # index (0 to N_sources-1) to the index of the group and the source
        # index within the group for fast access.
        self._sidx_to_gidx_gsidx_map_arr = np.empty((0, 2), dtype=np.int32)

        # Add source hypo groups if specified.
        if src_hypo_groups is not None:
            if isinstance(src_hypo_groups, SourceHypoGroup):
                src_hypo_groups = [src_hypo_groups]
            if not issequenceof(src_hypo_groups, SourceHypoGroup):
                raise TypeError(
                    'The src_hypo_groups argument must be an instance of '
                    'SourceHypoGroup, or a sequence of SourceHypoGroup '
                    'instances!')
            for shg in src_hypo_groups:
                self._shg_list.append(shg)
                self._extend_sidx_to_gidx_gsidx_map_arr(shg)

    @property
    def source_list(self):
        """The list of defined SourceModel instances.
        """
        source_list = []
        for shg in self._shg_list:
            source_list += shg.source_list
        return source_list

    @property
    def n_sources(self):
        """(read-only) The total number of sources defined in all source groups.
        """
        return self._sidx_to_gidx_gsidx_map_arr.shape[0]

    @property
    def n_src_hypo_groups(self):
        """DEPRICATED: Use n_shgs instead.
        The number of defined source hypothesis groups.
        """
        return len(self._shg_list)

    @property
    def n_shgs(self):
        """The number of defined source hypothesis groups.
        """
        return len(self._shg_list)

    @property
    def shg_list(self):
        """(read-only) The list of source hypothesis groups, i.e.
        SourceHypoGroup instances.
        """
        return self._shg_list

    def __str__(self):
        """Pretty string representation of this SourceHypoGroupManager.
        """
        s = f'{classname(self)}\n'

        s1 = 'Source Hypothesis Groups:'
        for (idx, shg) in enumerate(self._shg_list):
            s1 += '\n'
            s1 += add_leading_text_line_padding(INDENTATION_WIDTH, f'{idx}: {shg}')

        s += add_leading_text_line_padding(INDENTATION_WIDTH, s1)

        return s

    def _extend_sidx_to_gidx_gsidx_map_arr(self, shg):
        """Extends the source index to (group index, group source index) map
        array by one source hypo group.

        Parameters
        ----------
        shg : SourceHypoGroup instance
            The SourceHypoGroup instance for which the map array should get
            extented.
        """
        arr = np.empty((shg.n_sources, 2), dtype=np.int32)
        arr[:, 0] = self.n_src_hypo_groups-1  # Group index.
        arr[:, 1] = np.arange(shg.n_sources)  # Group source index.
        self._sidx_to_gidx_gsidx_map_arr = np.vstack(
            (self._sidx_to_gidx_gsidx_map_arr, arr))

    def create_source_hypo_group(
            self,
            sources,
            fluxmodel,
            detsigyield_builders,
            sig_gen_method=None):
        """Creates and adds a source hypothesis group to this source hypothesis
        group manager. A source hypothesis group shares sources of the same
        source model with the same flux model and hence the same detector signal
        yield and signal generation methods.

        Parameters
        ----------
        sources : SourceModel | sequence of SourceModel
            The source or sequence of sources that define the source group.
        fluxmodel : instance of FluxModel
            The FluxModel instance that applies to the list of sources of the
            group.
        detsigyield_builders : sequence of DetSigYieldBuilder instances
            The sequence of detector signal yield builder instances,
            which should be used to create the detector signal
            yield for the sources of this group. Each element is the
            detector signal yield builder for the particular dataset, if
            several datasets are used. If this list contains only one builder,
            it should be used for all datasets.
        sig_gen_method : instance of SignalGenerationMethod | None
            The SignalGenerationMethod instance that implements the detector
            and source hypothesis specific signal generation.
            It can be set to None which means no signal can be generated.
        """
        # Create the source group.
        group = SourceHypoGroup(
            sources=sources,
            fluxmodel=fluxmodel,
            detsigyield_builders=detsigyield_builders,
            sig_gen_method=sig_gen_method)

        # Add the group.
        self._shg_list.append(group)

        # Extend the source index to (group index, group source index) map
        # array.
        self._extend_sidx_to_gidx_gsidx_map_arr(group)

    def get_fluxmodel_by_src_idx(self, src_idx):
        """Retrieves the FluxModel instance for the source specified by its
        source index.

        Parameters
        ----------
        src_idx : int
            The index of the source, which must be in the range
            [0, N_sources-1].

        Returns
        -------
        fluxmodel : instance of FluxModel
            The FluxModel instance that applies to the specified source.
        """
        gidx = self._sidx_to_gidx_gsidx_map_arr[src_idx, 0]
        return self._shg_list[gidx]._fluxmodel

    def get_detsigyield_builder_list_by_src_idx(self, src_idx):
        """Retrieves the list of DetSigYieldBuilder instances for the source
        specified by its source index.

        Parameters
        ----------
        src_idx : int
            The index of the source, which must be in the range
            [0, N_sources-1].

        Returns
        -------
        detsigyield_builder_list : list of DetSigYieldBuilder instances
            The list of DetSigYieldBuilder instances that apply to the
            specified source.
        """
        gidx = self._sidx_to_gidx_gsidx_map_arr[src_idx, 0]
        return self._shg_list[gidx]._detsigyield_builder_list

    def get_src_mask_of_shg(self, shg_idx):
        """Creates a source mask for the sources of the ``shg_idx`` th source
        hypothesis group.

        Parameters
        ----------
        shg_idx : int
            The index of the source hypothesis group.

        Returns
        -------
        src_mask : instance of numpy ndarray
            The (N_sources,)-shaped numpy ndarray of bool holding the mask for
            selecting the sources of the given source hypothesis group.
        """
        return (self._sidx_to_gidx_gsidx_map_arr[:, 0] == shg_idx)

    def get_src_idxs_of_shg(self, shg_idx):
        """Creates an array of indices of sources that belong to the given
        source hypothesis group.

        Parameters
        ----------
        shg_idx : int
            The index of the source hypothesis group.

        Returns
        -------
        src_idxs : instance of numpy ndarray
            The numpy ndarray of int holding the indices of the sources that
            belong to the given source hypothesis group.
        """
        src_idxs = np.arange(self.n_sources)[self.get_src_mask_of_shg(shg_idx)]

        return src_idxs
